"""
Created on 24 Jan 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"sample": {"rec": "2024-01-24T14:32:35Z", "tag": "scs-be2-3", "ver": 2.0, "src": "N3",
"val": {"per": 4.1, "pm1": 3.4, "pm2p5": 5.8, "pm10": 6.1,
"bin": [276, 175, 80, 13, 11, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
"mtf1": 27, "mtf3": 29, "mtf5": 40, "mtf7": 0, "sfr": 5.61,
"sht": {"hmd": 42.6, "tmp": 23.8}}},
"ext-sht": {"hmd": 46.6, "tmp": 24.9},
"slopes": {"ext-t": 0.1, "ext-rh": 0.2, "opc-t": 0.3, "opc-rh": 0.4}}
"""

from collections import OrderedDict

from scs_core.data.datum import Datum

from scs_core.model.pmx.pmx_request import PMxRequest as AbstractPMxRequest


# --------------------------------------------------------------------------------------------------------------------

class PMxRequest(AbstractPMxRequest):
    """
    classdocs
    """

    __SLOPE_PERIOD = 900                # 15 minutes in seconds

    @classmethod
    def slope_tally(cls, schedule_duration):
        periods = int(round(cls.__SLOPE_PERIOD / schedule_duration))
        tally = 2 if periods < 2 else periods

        return tally


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sample, sht_datum, ext_t_slope, ext_rh_slope, opc_t_slope, opc_rh_slope):
        """
        Constructor
        """
        self.__sample = sample                                                  # ParticulatesSample
        self.__sht_datum = sht_datum                                            # SHTDatum

        self.__ext_t_slope = Datum.float(ext_t_slope, ndigits=6)                # float or None
        self.__ext_rh_slope = Datum.float(ext_rh_slope, ndigits=6)              # float or None

        self.__opc_t_slope = Datum.float(opc_t_slope, ndigits=6)                # float or None
        self.__opc_rh_slope = Datum.float(opc_rh_slope, ndigits=6)              # float or None


    # ----------------------------------------------------------------------------------------------------------------

    def is_compatible(self, group):
        return self.sample.src == group.opc_version()


    def is_zero(self):
        return self.sample.opc_datum.is_zero()


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['sample'] = self.sample
        jdict['ext-sht'] = self.sht_datum

        jdict['slopes'] = {
            'ext-t': self.ext_t_slope,
            'ext-rh': self.ext_rh_slope,
            'opc-t': self.opc_t_slope,
            'opc-rh': self.opc_rh_slope
        }

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def sample(self):
        return self.__sample


    @property
    def sht_datum(self):
        return self.__sht_datum


    @property
    def ext_t_slope(self):
        return self.__ext_t_slope


    @property
    def ext_rh_slope(self):
        return self.__ext_rh_slope


    @property
    def opc_t_slope(self):
        return self.__opc_t_slope


    @property
    def opc_rh_slope(self):
        return self.__opc_rh_slope


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PMxRequest(o2):{sample:%s, sht_datum:%s, ext_t_slope:%s, ext_rh_slope:%s, " \
               "opc_t_slope:%s, opc_rh_slope:%s}" %  \
               (self.sample, self.sht_datum, self.ext_t_slope, self.ext_rh_slope,
                self.opc_t_slope, self.opc_rh_slope)



"""
Created on 30 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Alphasense A4 electrochemical sensor
"""

from scs_core.gas.sensor import Sensor

from scs_core.gas.a4.a4_calibrated_datum import A4Calibrator
from scs_core.gas.a4.a4_datum import A4Datum
from scs_core.gas.a4.a4_temp_comp import A4TempComp


# --------------------------------------------------------------------------------------------------------------------

class A4(Sensor):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def init(cls):
        cls.SENSORS[cls.CODE_CO] =      A4(cls.CODE_CO,     'CO A4',    'CO',   3)
        cls.SENSORS[cls.CODE_H2S] =     A4(cls.CODE_H2S,    'H2SA4',    'H2S',  3)
        cls.SENSORS[cls.CODE_NO] =      A4(cls.CODE_NO,     'NO A4',    'NO',   3)
        cls.SENSORS[cls.CODE_NO2] =     A4(cls.CODE_NO2,    'NO2A43F',  'NO2',  3)      # was NOGA4
        cls.SENSORS[cls.CODE_OX] =      A4(cls.CODE_OX,     'OXA431',   'Ox',   3)      # was OXGA4, ref in A4Calib
        cls.SENSORS[cls.CODE_SO2] =     A4(cls.CODE_SO2,    'SO2A4',    'SO2',  3)

        cls.SENSORS[cls.CODE_VOCe] =    A4(cls.CODE_VOCe,   'VOCA4',    'VOCe', 3)

        cls.SENSORS[cls.CODE_TEST_1] =  A4(cls.CODE_TEST_1, 'TEST',     'SN1',  3)
        cls.SENSORS[cls.CODE_TEST_2] =  A4(cls.CODE_TEST_2, 'TEST',     'SN2',  3)
        cls.SENSORS[cls.CODE_TEST_3] =  A4(cls.CODE_TEST_3, 'TEST',     'SN3',  3)
        cls.SENSORS[cls.CODE_TEST_4] =  A4(cls.CODE_TEST_4, 'TEST',     'SN4',  3)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, sensor_code, sensor_type, gas_name, adc_gain_index):
        """
        Constructor
        """
        Sensor.__init__(self, sensor_code, sensor_type, gas_name, adc_gain_index)

        self.__tc = A4TempComp.find(sensor_code)


    # ----------------------------------------------------------------------------------------------------------------

    def sample(self, afe, temp, sensor_index, no2_sample=None):
        we_v, ae_v = afe.sample_raw_wrk_aux(sensor_index, self.adc_gain_index)

        if self.has_no2_cross_sensitivity():
            if no2_sample is None:
                raise ValueError("NO2 sample required, but none given.")

            return self.datum(temp, we_v, ae_v, no2_cnc=no2_sample.cnc)

        return self.datum(temp, we_v, ae_v)


    def datum(self, temp, we_v, ae_v, no2_cnc=None):
        datum = A4Datum.construct(self.calib, self.baseline, self.__tc, temp, we_v, ae_v, no2_cnc=no2_cnc)

        if self.calibrator is None:
            return datum

        return self.calibrator.calibrate(datum)     # no2_cnc is not supplied here - wait for the exegesis NO2 value


    def null_datum(self):
        return A4Datum(None, None)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tc(self):
        return self.__tc


    @property
    def calib(self):
        return self._calib


    @calib.setter
    def calib(self, calib):
        self._calib = calib
        self._calibrator = A4Calibrator(calib)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4:{sensor_code:%s, sensor_type:%s, gas_name:%s, calib:%s, baseline:%s, tc:%s, calibrator:%s}" %  \
               (self.sensor_code, self.sensor_type, self.gas_name, self.calib, self.baseline, self.tc, self.calibrator)



"""
Created on 29 May 2017

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

A JSONable wrapper for datetime.timedelta.

Warning: for the accessor methods, only days() carries the sign of the delta.
This is consistent with datetime.timedelta. You have been warned.
"""

import re
import sys

from datetime import timedelta

from scs_core.data.json import JSONable
from scs_core.sys.exception_report import ExceptionReport


# --------------------------------------------------------------------------------------------------------------------

class Timedelta(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_flag(cls, report):
        if report is None:
            return None

        # command-line utility flag...
        match = re.match(r'^(((\d+)-)?(\d{1,2}):)?(\d{1,2})?(:(\d{1,2}))?$', report)

        if match is None:
            return None

        fields = match.groups()

        days = 0 if fields[2] is None else int(fields[2])
        hours = 0 if fields[3] is None else int(fields[3])
        minutes = 0 if fields[4] is None else int(fields[4])
        seconds = 0 if fields[6] is None else int(fields[6])

        if days is None and hours is None and minutes is None and seconds is None:
            return None

        return cls(days=days, hours=hours, minutes=minutes, seconds=seconds)


    @classmethod
    def construct_from_ps_time_report(cls, report):
        if report is None:
            return None

        # CPU time...
        match = re.match(r'(\d+)?:?(\d+):(\d+)\.?(\d{2})?', report)

        if match is None:
            return None

        fields = match.groups()

        hours = 0 if fields[0] is None else int(fields[0])
        minutes = int(fields[1])
        seconds = int(fields[2])
        milliseconds = 0 if fields[3] is None else int(fields[3]) * 10

        return Timedelta(hours=hours, minutes=minutes, seconds=seconds, milliseconds=milliseconds)


    @classmethod
    def construct_from_ps_elapsed_report(cls, report):
        if report is None:
            return None

        # elapsed time...
        match = re.match(r'(\d+)?(-)?(\d+)?:?(\d+):(\d+)', report)

        if match is None:
            return None

        fields = match.groups()

        if fields[1] == "-":
            days = int(fields[0])
            hours = 0 if fields[2] is None else int(fields[2])
        else:
            days = 0
            hours = 0 if fields[0] is None else int(fields[0])

        minutes = int(fields[3])
        seconds = int(fields[4])

        return cls(days=days, hours=hours, minutes=minutes, seconds=seconds)


    @classmethod
    def construct_from_uptime_report(cls, report):
        if report is None:
            return None

        # uptime...
        match = re.match(r'.*up (\d+)?\s*(day)?s?,?\s*(\d+)?\s*(min)?s?,?\s*(\d+)?:?(\d+)?,',
                         report)

        if match:
            fields = match.groups()

            try:
                if fields[1] == 'day' and fields[3] == 'min':
                    days = int(fields[0])
                    hours = 0
                    minutes = int(fields[2])

                elif fields[1] == 'day' and fields[3] is None:
                    days = int(fields[0])
                    hours = 0 if fields[2] is None else int(fields[2])
                    minutes = 0 if fields[5] is None else int(fields[5])

                elif fields[1] is None and fields[3] == 'min':
                    days = 0
                    hours = 0
                    minutes = int(fields[0])

                else:
                    days = 0
                    hours = int(fields[0] if fields[0] is not None else fields[2])
                    minutes = int(fields[5])

            except TypeError as ex:
                print('Timedelta: unparsable:[%s]' % report.strip(), file=sys.stderr)
                print('Timedelta: fields:%s' % str(fields), file=sys.stderr)
                print(ExceptionReport.construct(ex), file=sys.stderr)
                return None

            return cls(days=days, hours=hours, minutes=minutes)


    @classmethod
    def construct_from_dst_report(cls, report):
        if report is None:
            return None

        # CPU time...
        match = re.match(r'(\d+):(\d+):(\d+)', report)

        if match is None:
            return None

        fields = match.groups()

        hours = int(fields[0])
        minutes = int(fields[1])
        seconds = int(fields[2])

        return Timedelta(hours=hours, minutes=minutes, seconds=seconds)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if jdict is None:
            return None

        match = re.match(r'(\d+)-(\d{2}):(\d{2}):(\d{2})(.(\d{3}))?', jdict)

        if match is None:
            return None

        fields = match.groups()

        days = int(fields[0])
        hours = int(fields[1])
        minutes = int(fields[2])
        seconds = int(fields[3])
        milliseconds = 0 if fields[5] is None else int(fields[5])

        return cls(days=days, hours=hours, minutes=minutes, seconds=seconds, milliseconds=milliseconds)


    @classmethod
    def construct(cls, td: timedelta):
        return cls(seconds=int(round(td.total_seconds())))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, weeks=0, days=0, hours=0, minutes=0, seconds=0, milliseconds=0, microseconds=0):
        """
        Constructor
        """
        self.__td = timedelta(days=days, seconds=seconds, microseconds=microseconds, milliseconds=milliseconds,
                              minutes=minutes, hours=hours, weeks=weeks)


    def __eq__(self, other):
        try:
            return self.delta == other.delta

        except (TypeError, AttributeError):
            return False


    def __ge__(self, other):
        return self.delta >= other.delta


    def __gt__(self, other):
        return self.delta > other.delta


    def __le__(self, other):
        return self.delta <= other.delta


    def __lt__(self, other):
        return self.delta < other.delta


    def __add__(self, other):
        seconds = self.total_seconds() + other.total_seconds()
        return Timedelta(seconds=seconds)


    def __sub__(self, other):
        seconds = self.total_seconds() - other.total_seconds()
        return Timedelta(seconds=seconds)


    def __truediv__(self, other):
        seconds = self.total_seconds() / other
        return Timedelta(seconds=seconds)


    def __mul__(self, other):
        seconds = self.total_seconds() * other
        return Timedelta(seconds=seconds)


    __rmul__ = __mul__


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        json = "%02d-%02d:%02d:%02d" % (self.days, self.hours, self.minutes, self.seconds)

        if self.milliseconds != 0:
            json += ".%03d" % self.milliseconds

        return json


    # ----------------------------------------------------------------------------------------------------------------

    def total_seconds(self):
        return self.__td.total_seconds()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def delta(self):
        return self.__td


    @property
    def days(self):
        return self.__td.days


    @property
    def hours(self):
        return self.__td.seconds // 3600


    @property
    def minutes(self):
        return self.__td.seconds % 3600 // 60


    @property
    def seconds(self):
        return self.__td.seconds % 60


    @property
    def milliseconds(self):
        return self.__td.microseconds // 1000


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Timedelta:{%s}" % self.__td



"""
Created on 5 Apr 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.security.cognito_device import CognitoDeviceIdentity
from scs_core.aws.security.cognito_device_manager import Endpoint


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find_all(self, token):
        url = Endpoint.url('retrieve', 'all')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return tuple(CognitoDeviceIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def find_by_tag(self, token, tag):
        url = Endpoint.url('retrieve', 'in')
        payload = {"username": tag}

        response = requests.get(url, headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return tuple(CognitoDeviceIdentity.construct_from_jdict(jdict) for jdict in response.json())


    def get_by_tag(self, token, tag):
        url = Endpoint.url('retrieve', 'exact')
        payload = {"username": tag}

        response = requests.get(url, headers=self._token_headers(token), json=payload)
        self._check_response(response)

        return CognitoDeviceIdentity.construct_from_jdict(response.json())


# --------------------------------------------------------------------------------------------------------------------

class CognitoDeviceIntrospector(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find_self(self, token):
        url = Endpoint.url('self')

        response = requests.get(url, headers=self._token_headers(token))
        self._check_response(response)

        return CognitoDeviceIdentity.construct_from_jdict(response.json())



"""
Created on 8 Aug 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.client_traffic.client_traffic_intercourse import ClientTrafficResponse
from scs_core.aws.config.endpoint import APIEndpoint


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CliTrfAPI/ClientTraffic',
                   'https://tduyom430a.execute-api.us-west-2.amazonaws.com/default/ClientTraffic')


# --------------------------------------------------------------------------------------------------------------------

class ClientTrafficFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def find_for_users(self, token, request):
        url = Endpoint.url('users')

        for item in self._get_blocks(url, token, ClientTrafficResponse, payload=request):
            yield item


    def find_for_organisations(self, token, request):
        url = Endpoint.url('organisations')

        for item in self._get_blocks(url, token, ClientTrafficResponse, payload=request):
            yield item


    def find_for_organisations_users(self, token, request):
        url = Endpoint.url('organisations-users')

        for item in self._get_blocks(url, token, ClientTrafficResponse, payload=request):
            yield item


    def find_for_endpoint(self, token, request):
        url = Endpoint.url('endpoint')

        for item in self._get_blocks(url, token, ClientTrafficResponse, payload=request):
            yield item



"""
Created on 21 Sep 2020

@author: Jade Page (jade.page@southcoastscience.com)

The aws_group_configurator is a class used by aws_group_setup to collate information and make requests to
the amazon greengrass API. It is dependent on aws_json_reader to collect information.

example document
{"group-name": "scs-bbe-651-group", "time-initiated": "2021-09-21T13:00:31Z", "unix-group": 987, "ml": "oE1"}
"""

import grp
import json
import os

from collections import OrderedDict

from scs_core.aws.config.project import Project
from scs_core.aws.greengrass.aws_group import AWSGroup
from scs_core.aws.greengrass.gg_errors import ProjectMissingError

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import PersistentJSONable
from scs_core.data.path_dict import PathDict

from scs_core.model.model_conf import ModelConf

from scs_core.sys.filesystem import Filesystem
from scs_core.sys.logging import Logging
from scs_core.sys.system_id import SystemID


# --------------------------------------------------------------------------------------------------------------------

class AWSGroupConfiguration(PersistentJSONable):
    """
    classdocs
    """
    __CATALOGUE_NAME = 'templates'

    @classmethod
    def catalogue_location(cls):
        return os.path.join(os.path.dirname(os.path.realpath(__file__)), cls.__CATALOGUE_NAME)


    @classmethod
    def templates(cls):
        return tuple(item.name for item in Filesystem.ls(cls.catalogue_location()) if item.is_directory)


    @classmethod
    def read(cls, catalogue, file):
        abs_path = os.path.join(cls.catalogue_location(), catalogue, file)

        with open(abs_path) as f:
            return json.load(f)


    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "aws_group_config.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        group_name = jdict.get('group-name')
        init_time = LocalizedDatetime.construct_from_iso8601(jdict.get('time-initiated'))
        ml = jdict.get('ml')
        unix_group = jdict.get('unix-group')

        if isinstance(ml, bool):
            ml = 'vB.1' if ml else 'm0'

        return cls(group_name, init_time, unix_group=unix_group, ml=ml)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, group_name, init_time, unix_group=None, ml='m0'):
        """
        Constructor
        """
        super().__init__()

        if unix_group is None:
            try:
                group_info = grp.getgrnam('ggc_user')
                unix_group = group_info[2]
            except KeyError:
                raise KeyError('GGC_USER')

        self.__group_name = group_name                  # string
        self.__init_time = init_time                    # LocalisedDatetime
        self.__unix_group = unix_group                  # int
        self.__ml = ml                                  # string (was bool)


    def __eq__(self, other):
        try:
            return self.group_name == other.group_name and self.init_time == other.init_time and \
                   self.unix_group == other.unix_group and self.ml == other.ml

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        if self.__init_time is None:
            self.__init_time = LocalizedDatetime.now()

        super().save(manager, encryption_key=encryption_key)


    # ----------------------------------------------------------------------------------------------------------------

    def configuration_is_mismatched(self, gas_model_conf, pmx_model_conf):
        gg_ml_template = ModelConf.gg_ml_template(gas_model_conf, pmx_model_conf)       # may raise ValueError

        if gg_ml_template is not None and self.ml is not None:
            return gg_ml_template != self.ml

        return False


    def configurator(self, client):
        return AWSGroupConfigurator(self, client)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['group-name'] = self.group_name
        jdict["time-initiated"] = self.init_time
        jdict['unix-group'] = self.unix_group
        jdict['ml'] = self.ml

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def group_name(self):
        return self.__group_name


    @property
    def init_time(self):
        return self.__init_time


    @property
    def unix_group(self):
        return self.__unix_group


    @property
    def ml(self):
        return self.__ml


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AWSGroupConfiguration:{group_name:%s, init_time:%s, unix_group:%s, ml:%s}" % \
               (self.group_name, self.init_time, self.unix_group, self.ml)


# --------------------------------------------------------------------------------------------------------------------

class AWSGroupConfigurator(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, config: AWSGroupConfiguration, client):
        """
        Constructor
        """
        self.__logger = Logging.getLogger()

        self.__config = config                      # AWSGroupConfiguration
        self.__client = client                      # Client
        self.__aws_info = PathDict()                # PathDict


    # ----------------------------------------------------------------------------------------------------------------

    def collect_information(self, host):
        aws_json_reader = AWSGroup(self.config.group_name, self.__client)
        aws_json_reader.get_group_info_from_name()
        aws_json_reader.get_group_arns()
        self.__aws_info.append("GroupID", aws_json_reader.retrieve_node("GroupID"))
        self.__aws_info.append("GroupVersionID", aws_json_reader.retrieve_node("GroupLatestVersionID"))
        self.__aws_info.append("CoreDefinitionARN", aws_json_reader.retrieve_node("CoreDefinitionVersionArn"))
        self.__aws_info.append("FunctionDefinitionARN", aws_json_reader.retrieve_node("FunctionDefinitionVersionArn"))
        self.__aws_info.append("ResourceDefinitionARN", aws_json_reader.retrieve_node("ResourceDefinitionVersionArn"))
        self.__aws_info.append("SubscriptionDefinitionARN",
                               aws_json_reader.retrieve_node("SubscriptionDefinitionVersionArn"))

        temp = SystemID.load(host)
        device = temp.topic_label()
        temp_json = temp.as_json()
        self.__aws_info.append("SystemID", temp_json['system-sn'])  # can only run on a setup device

        project = Project.load(host)

        if not project:
            raise ProjectMissingError()

        else:
            project_json = project.as_json()
            device_path = project_json["device-path"] + "/" + str(device)
            self.__aws_info.append("LocationPath", project_json["location-path"])
            self.__aws_info.append("DevicePath", device_path)


    # ----------------------------------------------------------------------------------------------------------------

    def define_aws_group_resources(self, host):
        # Setup default JSON
        r_data = AWSGroupConfiguration.read(self.config.ml, 'gg_resources.json')

        # Edit JSON for device
        system_id = str(self.__aws_info.node("SystemID"))
        temp = host.home_path()
        group_owner_name = temp.split("/")[2]
        scs_path = host.scs_path()
        r_data["Name"] = ("Resources-" + system_id)
        r_data_resources = r_data["InitialVersion"]["Resources"]

        r_data_resources[0]["Id"] = (system_id + "-data-volume")
        r_data_resources[0]["ResourceDataContainer"]["LocalVolumeResourceData"][
            "GroupOwnerSetting"]["AutoAddGroupOwner"] = False
        r_data_resources[0]["ResourceDataContainer"]["LocalVolumeResourceData"][
            "SourcePath"] = scs_path
        r_data_resources[0]["ResourceDataContainer"]["LocalVolumeResourceData"][
            "GroupOwnerSetting"]["GroupOwner"] = group_owner_name

        if self.config.ml:
            r_data_resources[1]["Id"] = (system_id + "-ml-pm1")
            r_data_resources[2]["Id"] = (system_id + "-ml-pm2p5")
            r_data_resources[3]["Id"] = (system_id + "-ml-pm10")

            try:
                r_data_resources[4]["Id"] = (system_id + "-ml-no2")
            except IndexError:
                pass

            try:
                r_data_resources[5]["Id"] = (system_id + "-ml-no")
            except IndexError:
                pass

            try:
                r_data_resources[6]["Id"] = (system_id + "-ml-o3")
            except IndexError:
                pass

            try:
                r_data_resources[7]["Id"] = (system_id + "-ml-so2")
            except IndexError:
                pass

        # Send request
        self.__logger.info("Creating resource definition")
        self.__logger.info(r_data)

        response = self.__client.create_resource_definition(InitialVersion=r_data["InitialVersion"])
        self.__aws_info.append("NewResourceARN", response["LatestVersionArn"])

        self.__logger.info(response)


    # ----------------------------------------------------------------------------------------------------------------

    def define_aws_group_functions(self):
        # Get template JSON
        f_data = AWSGroupConfiguration.read(self.config.ml, 'gg_functions.json')

        # Update JSON for device
        system_id = str(self.__aws_info.node("SystemID"))
        data_volume_name = (system_id + "-data-volume")
        f_data_functions = f_data["InitialVersion"]["Functions"]

        f_data_functions[0]["Id"] = (system_id + "-ControlSubscriber")
        f_data_functions[1]["Id"] = (system_id + "-TopicPublisher")

        f_data_functions[0]["FunctionConfiguration"]["Environment"]["ResourceAccessPolicies"][0][
            "ResourceId"] = data_volume_name
        f_data_functions[1]["FunctionConfiguration"]["Environment"]["ResourceAccessPolicies"][0][
            "ResourceId"] = data_volume_name

        if self.config.ml:
            f_data_functions[2]["Id"] = (system_id + "-PMxInference")
            f_data_functions[2]["FunctionConfiguration"]["Environment"]["ResourceAccessPolicies"][
                0]["ResourceId"] = data_volume_name
            f_data_functions[2]["FunctionConfiguration"]["Environment"]["ResourceAccessPolicies"][
                1]["ResourceId"] = (system_id + "-ml-pm1")
            f_data_functions[2]["FunctionConfiguration"]["Environment"]["ResourceAccessPolicies"][
                2]["ResourceId"] = (system_id + "-ml-pm2p5")
            f_data_functions[2]["FunctionConfiguration"]["Environment"]["ResourceAccessPolicies"][
                3]["ResourceId"] = (system_id + "-ml-pm10")

            try:
                f_data_functions[3]["Id"] = (system_id + "-GasInference")
            except IndexError:
                pass

            try:
                f_data_functions[3]["FunctionConfiguration"]["Environment"][
                    "ResourceAccessPolicies"][0]["ResourceId"] = data_volume_name
            except IndexError:
                pass

            try:
                f_data_functions[3]["FunctionConfiguration"]["Environment"][
                    "ResourceAccessPolicies"][1]["ResourceId"] = (system_id + "-ml-no2")
            except IndexError:
                pass

            try:
                f_data_functions[3]["FunctionConfiguration"]["Environment"][
                    "ResourceAccessPolicies"][2]["ResourceId"] = (system_id + "-ml-no")
            except IndexError:
                pass

            try:
                f_data_functions[3]["FunctionConfiguration"]["Environment"][
                    "ResourceAccessPolicies"][3]["ResourceId"] = (system_id + "-ml-o3")
            except IndexError:
                pass

            try:
                f_data_functions[3]["FunctionConfiguration"]["Environment"][
                    "ResourceAccessPolicies"][4]["ResourceId"] = (system_id + "-ml-so2")
            except IndexError:
                pass

        self.__logger.info("Creating function definition")
        self.__logger.info(f_data)

        # Create request
        response = self.__client.create_function_definition(InitialVersion=f_data["InitialVersion"])
        self.__aws_info.append("NewFunctionARN", response["LatestVersionArn"])

        self.__logger.info(response)


    # ----------------------------------------------------------------------------------------------------------------

    def define_aws_group_subscriptions(self):
        s_data = AWSGroupConfiguration.read('m0', 'gg_subscriptions.json')

        # Edit for device
        system_id = str(self.__aws_info.node("SystemID"))
        s_data_subs = s_data["InitialVersion"]["Subscriptions"]
        s_data_subs[0]["Id"] = (system_id + "-particulates-subscription")
        s_data_subs[1]["Id"] = (system_id + "-control-from-cloud-subscription")
        s_data_subs[2]["Id"] = (system_id + "-climate-subscription")
        s_data_subs[3]["Id"] = (system_id + "-status-subscription")
        s_data_subs[4]["Id"] = (system_id + "-control-to-cloud-subscription")
        s_data_subs[5]["Id"] = (system_id + "-gases-subscription")

        s_data_subs[0]["Subject"] = (self.__aws_info.node("LocationPath") + "/particulates")
        s_data_subs[1]["Subject"] = (self.__aws_info.node("DevicePath") + "/control")
        s_data_subs[2]["Subject"] = (self.__aws_info.node("LocationPath") + "/climate")
        s_data_subs[3]["Subject"] = (self.__aws_info.node("DevicePath") + "/status")
        s_data_subs[4]["Subject"] = (self.__aws_info.node("DevicePath") + "/control")
        s_data_subs[5]["Subject"] = (self.__aws_info.node("LocationPath") + "/gases")

        self.__logger.info("Creating sub definition")
        self.__logger.info(s_data)

        # Send request
        response = self.__client.create_subscription_definition(InitialVersion=s_data["InitialVersion"])
        self.__aws_info.append("NewSubscriptionARN", response["LatestVersionArn"])

        self.__logger.info(response)


    def define_aws_logger(self):
        l_data = AWSGroupConfiguration.read('m0', 'gg_logger.json')

        system_id = str(self.__aws_info.node("SystemID"))
        l_data["InitialVersion"]["Loggers"][0]["Id"] = (system_id + "-lambda-logger")

        self.__logger.info("Creating logger definition")
        self.__logger.info(l_data)

        response = self.__client.create_logger_definition(InitialVersion=l_data["InitialVersion"])
        self.__aws_info.append("NewLoggerARN", response["LatestVersionArn"])

        self.__logger.info(response)


    def create_aws_group_definition(self):
        self.__logger.info("Creating group definition")
        self.__logger.info(self.__aws_info.node("GroupID"))

        response = self.__client.create_group_version(
            CoreDefinitionVersionArn=self.__aws_info.node("CoreDefinitionARN"),
            FunctionDefinitionVersionArn=self.__aws_info.node("NewFunctionARN"),
            GroupId=self.__aws_info.node("GroupID"),
            ResourceDefinitionVersionArn=self.__aws_info.node("NewResourceARN"),
            SubscriptionDefinitionVersionArn=self.__aws_info.node("NewSubscriptionARN"),
            # LoggerDefinitionVersionArn=self.__aws_info.node("NewLoggerARN"),
        )

        self.__logger.info(response)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def config(self):
        return self.__config


    @property
    def client(self):
        return self.__client


    @property
    def aws_info(self):
        return self.__aws_info


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "AWSGroupConfigurator:{config:%s, client:%s, aws_info:%s}" % \
               (self.config, self.client, self.aws_info)



"""
Created on 25 Jan 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""


# --------------------------------------------------------------------------------------------------------------------

class ModelMap(object):
    """
    classdocs
    """

    __MAPS =       {}

    @classmethod
    def init(cls):
        cls.__MAPS['oE.1'] = cls('oE.1', 'oE.1', 'oE.1', 'g0')
        cls.__MAPS['oM.2'] = cls('oM.2', 'oE.1', 'oPG.2', 'oP.2')
        cls.__MAPS['uE.1'] = cls('uE.1', 'uE.1', 'uE.1', 'g0')


    @classmethod
    def map(cls, name):
        return cls.__MAPS[name]                             # may raise KeyError


    @classmethod
    def names(cls):
        return sorted(cls.__MAPS.keys())


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, gas_compendium_group, pg_gg_ml_template, p_gg_ml_template):
        """
        Constructor
        """
        self.__name = name                                                  # string

        self.__gas_compendium_group = gas_compendium_group                  # string
        self.__pg_gg_ml_template = pg_gg_ml_template                        # string PMx + Gas
        self.__p_gg_ml_template = p_gg_ml_template                          # string PMx


    def __eq__(self, other):
        try:
            return self.name == other.name and self.gas_compendium_group == other.gas_compendium_group and \
                   self.pg_gg_ml_template == other.pg_gg_ml_template and self.p_gg_ml_template == other.p_gg_ml_template

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    @property
    def gas_compendium_group(self):
        return self.__gas_compendium_group


    @property
    def pg_gg_ml_template(self):
        return self.__pg_gg_ml_template


    @property
    def p_gg_ml_template(self):
        return self.__p_gg_ml_template


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ModelMap:{name:%s, gas_compendium_group:%s, pg_gg_ml_template:%s, p_gg_ml_template:%s}" %  \
               (self.name, self.gas_compendium_group, self.pg_gg_ml_template, self.p_gg_ml_template)



"""
Created on 20 Apr 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"email": "bruno.beloff@southcoastscience.com", "password": "XXX"}
"""

import sys
import termios

from collections import OrderedDict
from getpass import getpass

from scs_core.data.datum import Datum
from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class MonitorAuth(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "monitor_auth.json"

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def from_user(cls):
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)           # flush stdin
        except termios.error:
            pass

        print("Email address: ", end="", file=sys.stderr)
        email_address = input().strip()

        print("Password: ", end="", file=sys.stderr)
        password = getpass().strip()

        return cls(email_address, password)


    @staticmethod
    def password_from_user():
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)           # flush stdin
        except termios.error:
            pass

        return getpass().strip()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        email_address = jdict.get('email')
        password = jdict.get('password')

        return cls(email_address, password)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, email_address, password):
        """
        Constructor
        """
        super().__init__()

        self.__email_address = email_address                # String
        self.__password = password                          # String


    # ----------------------------------------------------------------------------------------------------------------

    def has_valid_email_address(self):
        return Datum.is_email_address(self.email_address)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['email'] = self.email_address
        jdict['password'] = self.password

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def email_address(self):
        return self.__email_address


    @property
    def password(self):
        return self.__password


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "MonitorAuth:{email_address:%s, password:%s}" % (self.email_address, self.password)



"""
Created on 20 Mar 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/2520893/how-to-flush-the-input-stream-in-python

example document:
{"email": "production@southcoastscience.com", "password": "###", "retrieval-password": "###"}
"""

import json
import os
import sys
import termios

from collections import OrderedDict
from getpass import getpass

from scs_core.aws.security.cognito_user import CognitoUserCredentials, CognitoUserIdentity

from scs_core.data.datum import Datum
from scs_core.data.json import MultiPersistentJSONable

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class CognitoClientCredentials(CognitoUserCredentials, MultiPersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "cognito_user_credentials.json"
    __ENV_PASSWORD = 'SCS_CREDENTIALS_RETRIEVAL'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def load_for_user(cls, host, name=None):
        logger = Logging.getLogger()

        if not cls.exists(host, name=name):
            logger.error("Cognito credentials not available.")
            return None

        try:
            password = os.environ[cls.__ENV_PASSWORD]
        except KeyError:
            password = cls.password_from_user()

        try:
            return cls.load(host, name=name, encryption_key=password)
        except (KeyError, ValueError):
            logger.error("incorrect password.")
            return None


    @classmethod
    def from_stdin(cls, name):
        line = sys.stdin.readline()
        jdict = json.loads(line)

        return cls.construct_from_jdict(jdict, name=name)


    @classmethod
    def from_user(cls, name, existing_email=None):
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)               # flush stdin
        except termios.error:
            pass

        if existing_email:
            email = existing_email

        else:
            print("Enter email address: ", end="", file=sys.stderr)
            email = input().strip()

        print("Enter password: ", end="", file=sys.stderr)
        password = input().strip()

        print("Enter retrieval password (RETURN for same): ", end="", file=sys.stderr)
        retrieval_password = input().strip()

        if not retrieval_password:
            retrieval_password = password

        return cls(name, email, password, retrieval_password)


    @staticmethod
    def password_from_user():
        try:
            termios.tcflush(sys.stdin, termios.TCIOFLUSH)               # flush stdin
        except termios.error:
            pass

        return getpass().strip()


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def persistence_location(cls, name):
        filename = cls.__FILENAME if name is None else '_'.join((name, cls.__FILENAME))

        return cls.aws_dir(), filename


    @classmethod
    def construct_from_jdict(cls, jdict, name=None, skeleton=False):
        if not jdict:
            return cls(None, None, None, None) if skeleton else None

        email = jdict.get('email')
        password = jdict.get('password')
        retrieval_password = jdict.get('retrieval-password')

        return cls(name, email, password, retrieval_password)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, email, password, retrieval_password):
        """
        Constructor
        """
        CognitoUserCredentials.__init__(self, email, password)
        MultiPersistentJSONable.__init__(self, name)

        self.__retrieval_password = retrieval_password                  # string


    # ----------------------------------------------------------------------------------------------------------------

    def ok(self):
        if not self.email or not self.password:
            return False

        if not Datum.is_email_address(self.email):
            return False

        if not CognitoUserIdentity.is_valid_password(self.password):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['email'] = self.email
        jdict['password'] = self.password
        jdict['retrieval-password'] = self.retrieval_password

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def retrieval_password(self):
        return self.__retrieval_password


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoClientCredentials:{name:%s, email:%s, password:%s, retrieval_password:%s}" %  \
               (self.name, self.email, self.password, self.retrieval_password)



"""
Created on 3 Mar 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://en.wikipedia.org/wiki/Standard_deviation


example documents:
Stats:
{"tag": "scs-bgx-619", "val": {"VOC": {"cnc": {"count": 4320, "min": 397.7, "mean": 448.2, "median": 446.1,
"max": 498.9, "var": 375.6, "stdev": 19.4, "stdev2": 38.8, "stdev3": 58.2}}}}

StatsAnalysis:
{"tag": "scs-bgx-570", "val": {"SO2": {"cnc": {"min": -3.6, "mean": 1.7, "median": 1.4, "max": 9.3,
"l3": -5.8, "l2": -3.4, "l1": -1.0, "u1": 3.8, "u2": 6.2, "u3": 8.6, "a1": 4.8, "a2": 9.6, "a3": 14.4}
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class Stats(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, values, prec=3):
        import statistics                               # late import

        minimum = round(min(values), prec)
        mean = round(statistics.mean(values), prec)
        median = round(statistics.median(values), prec)
        maximum = round(max(values), prec)

        variance = round(statistics.variance(values, xbar=mean), prec)

        stdev = round(statistics.stdev(values), prec)
        stdev2 = round(stdev * 2.0, prec)
        stdev3 = round(stdev * 3.0, prec)

        return cls(len(values), minimum, mean, median, maximum, variance, stdev, stdev2, stdev3)


    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        count = jdict.get('count')

        minimum = jdict.get('min')
        mean = jdict.get('mean')
        median = jdict.get('median')
        maximum = jdict.get('max')

        variance = jdict.get('var')

        stdev = jdict.get('stdev')
        stdev2 = jdict.get('stdev2')
        stdev3 = jdict.get('stdev3')

        return cls(count, minimum, mean, median, maximum, variance, stdev, stdev2, stdev3)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, count, minimum, mean, median, maximum, variance, stdev, stdev2, stdev3):
        """
        Constructor
        """
        self.__count = count                            # int

        self.__minimum = minimum                        # float
        self.__mean = mean                              # float
        self.__median = median                          # float
        self.__maximum = maximum                        # float

        self.__variance = variance                      # float

        self.__stdev = stdev                            # float
        self.__stdev2 = stdev2                          # float
        self.__stdev3 = stdev3                          # float


    def __len__(self):
        return self.count


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['count'] = self.count

        jdict['min'] = self.minimum
        jdict['mean'] = self.mean
        jdict['median'] = self.median
        jdict['max'] = self.maximum

        jdict['var'] = self.variance

        jdict['stdev'] = self.stdev
        jdict['stdev2'] = self.stdev2
        jdict['stdev3'] = self.stdev3

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def lower1(self):
        return self.median - self.stdev


    @property
    def lower2(self):
        return self.median - self.stdev2


    @property
    def lower3(self):
        return self.median - self.stdev3


    @property
    def upper1(self):
        return self.median + self.stdev


    @property
    def upper2(self):
        return self.median + self.stdev2


    @property
    def upper3(self):
        return self.median + self.stdev3


    @property
    def mm(self):
        return self.maximum - self.minimum


    @property
    def amp1(self):
        return self.upper1 - self.lower1


    @property
    def amp2(self):
        return self.upper2 - self.lower2


    @property
    def amp3(self):
        return self.upper3 - self.lower3


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def count(self):
        return self.__count


    @property
    def minimum(self):
        return self.__minimum


    @property
    def mean(self):
        return self.__mean


    @property
    def median(self):
        return self.__median


    @property
    def maximum(self):
        return self.__maximum


    @property
    def variance(self):
        return self.__variance


    @property
    def stdev(self):
        return self.__stdev


    @property
    def stdev2(self):
        return self.__stdev2


    @property
    def stdev3(self):
        return self.__stdev3


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Stats:{minimum:%s, mean:%s, median:%s, maximum:%s, variance:%s, stdev:%s, stdev2:%s, stdev3:%s}" %  \
               (self.minimum, self.mean, self.median, self.maximum, self.variance, self.stdev, self.stdev2, self.stdev3)


# --------------------------------------------------------------------------------------------------------------------

class StatsAnalysis(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, prec=1):
        if not jdict:
            return None

        minimum = jdict.get('min')
        mean = jdict.get('mean')
        median = jdict.get('median')
        maximum = jdict.get('max')

        lower3 = jdict.get('l3')
        lower2 = jdict.get('l2')
        lower1 = jdict.get('l1')

        upper1 = jdict.get('u1')
        upper2 = jdict.get('u2')
        upper3 = jdict.get('u3')

        mm = jdict.get('mm')

        amp1 = jdict.get('a1')
        amp2 = jdict.get('a2')
        amp3 = jdict.get('a3')

        return cls(minimum, mean, median, maximum, lower3, lower2, lower1, upper1, upper2, upper3, mm,
                   amp1, amp2, amp3, prec=prec)


    @classmethod
    def construct_from_stats_jdict(cls, jdict, prec=1):
        stats = Stats.construct_from_jdict(jdict)

        return cls(stats.minimum, stats.mean, stats.median, stats.maximum, stats.lower3, stats.lower2,
                   stats.lower1, stats.upper1, stats.upper2, stats.upper3, stats.mm,
                   stats.amp1, stats.amp2, stats.amp3, prec=prec)


    @classmethod
    def construct_from_stats(cls, stats, prec=1):
        return cls(stats.minimum, stats.mean, stats.median, stats.maximum, stats.lower3, stats.lower2,
                   stats.lower1, stats.upper1, stats.upper2, stats.upper3, stats.mm,
                   stats.amp1, stats.amp2, stats.amp3, prec=prec)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, minimum, mean, median, maximum, lower3, lower2, lower1, upper1, upper2, upper3, mm,
                 amp1, amp2, amp3, prec=1):
        """
        Constructor
        """
        self.__minimum = Datum.float(minimum, ndigits=prec)
        self.__mean = Datum.float(mean, ndigits=prec)
        self.__median = Datum.float(median, ndigits=prec)
        self.__maximum = Datum.float(maximum, ndigits=prec)

        self.__lower3 = Datum.float(lower3, ndigits=prec)
        self.__lower2 = Datum.float(lower2, ndigits=prec)
        self.__lower1 = Datum.float(lower1, ndigits=prec)

        self.__upper1 = Datum.float(upper1, ndigits=prec)
        self.__upper2 = Datum.float(upper2, ndigits=prec)
        self.__upper3 = Datum.float(upper3, ndigits=prec)

        self.__mm = Datum.float(mm, ndigits=prec)

        self.__amp1 = Datum.float(amp1, ndigits=prec)
        self.__amp2 = Datum.float(amp2, ndigits=prec)
        self.__amp3 = Datum.float(amp3, ndigits=prec)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['min'] = self.minimum
        jdict['mean'] = self.mean
        jdict['median'] = self.median
        jdict['max'] = self.maximum

        jdict['l3'] = self.lower3
        jdict['l2'] = self.lower2
        jdict['l1'] = self.lower1

        jdict['u1'] = self.upper1
        jdict['u2'] = self.upper2
        jdict['u3'] = self.upper3

        jdict['mm'] = self.mm

        jdict['a1'] = self.amp1
        jdict['a2'] = self.amp2
        jdict['a3'] = self.amp3

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def minimum(self):
        return self.__minimum


    @property
    def mean(self):
        return self.__mean


    @property
    def median(self):
        return self.__median


    @property
    def maximum(self):
        return self.__maximum


    @property
    def lower3(self):
        return self.__lower3


    @property
    def lower2(self):
        return self.__lower2


    @property
    def lower1(self):
        return self.__lower1


    @property
    def upper1(self):
        return self.__upper1


    @property
    def upper2(self):
        return self.__upper2


    @property
    def upper3(self):
        return self.__upper3


    @property
    def mm(self):
        return self.__mm


    @property
    def amp1(self):
        return self.__amp1


    @property
    def amp2(self):
        return self.__amp2


    @property
    def amp3(self):
        return self.__amp3


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "StatsAnalysis:{minimum:%s, mean:%s, median:%s, maximum:%s, lower3:%s, lower2:%s, lower1:%s, " \
               "upper1:%s, upper2:%s, upper3:%s, mm:%s, amp1:%s, amp2:%s, amp3:%s}" % \
               (self.minimum, self.mean, self.median, self.maximum, self.lower3, self.lower2, self.lower1,
                self.upper1, self.upper2, self.upper3, self.mm, self.amp1, self.amp2, self.amp3)



"""
Created on 28 Sep 2020

@author: Jade Page (jade.page@southcoastscience.com)
"""

from botocore.exceptions import ClientError
from collections import OrderedDict

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable
from scs_core.data.tokens import Tokens

from scs_core.sys.persistence_manager import PersistenceManager


# --------------------------------------------------------------------------------------------------------------------

class S3Manager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client, resource_client):
        """
        Constructor
        """
        self.__client = client
        self.__resource_client = resource_client


    # ----------------------------------------------------------------------------------------------------------------

    def list_buckets(self, full_details):
        response = self.__client.list_buckets()

        if 'Buckets' not in response:
            return

        for bucket in response['Buckets']:
            yield Bucket.construct(bucket) if full_details else bucket['Name']


    def list_objects(self, bucket, depth, full_details, prefix=None):
        prefix_tokens = None if prefix is None else Tokens.construct(prefix, '/')
        next_token = None
        summary = Summary.none()

        while True:
            response = self.__retrieve_objects(bucket, prefix, next_token)

            if not response or 'Contents' not in response:
                return

            for item in response['Contents']:
                key_tokens = Tokens.construct(item['Key'], '/')

                if prefix is None or key_tokens.startswith(prefix_tokens):
                    # full...
                    if full_details:
                        yield Object.construct(item)
                        continue

                    # minimal...
                    if depth is None:
                        yield item['Key']
                        continue

                    # summary...
                    path = key_tokens.path(depth=depth)
                    obj = Object.construct(item)

                    if path == summary.path:
                        summary.add(obj)
                        continue

                    if not summary.is_none():
                        yield summary

                    summary = Summary.new(path, obj)

            if 'NextContinuationToken' not in response:
                break

            next_token = response.get('NextContinuationToken')

        if summary.path is not None:
            yield summary


    def retrieve_from_bucket(self, bucket, key):
        response = self.__client.get_object(Bucket=bucket, Key=key)

        meta = response.get('ResponseMetadata')
        header = meta.get('HTTPHeaders')
        last_modified = LocalizedDatetime.construct_from_s3(header.get('last-modified'))

        content_body = response.get("Body")
        data = content_body.read().decode()

        return data, last_modified


    def upload_file_to_bucket(self, filepath, bucket, key):
        self.__resource_client.Bucket(bucket).upload_file(filepath, key)

        return self.head(bucket, key)


    def upload_bytes_to_bucket(self, body, bucket, key):
        self.__resource_client.Bucket(bucket).put_object(Body=body, Key=key)

        return self.head(bucket, key)


    def put_object(self, body, bucket, key):
        self.__client.put_object(Body=body, Bucket=bucket, Key=key)

        return self.head(bucket, key)


    def move_object(self, bucket, key, new_key):
        source = '/'.join((bucket, key))

        self.__client.copy_object(Bucket=bucket, CopySource=source, Key=new_key)
        self.__client.delete_object(Bucket=bucket, Key=key)

        return self.head(bucket, new_key)


    def delete_objects(self, bucket, prefix, excluded=None):
        excluded_tokens = Tokens.construct(excluded, '/')

        for key in self.list_objects(bucket, None, False, prefix=prefix):
            if excluded and Tokens.construct(key, '/').startswith(excluded_tokens):
                continue

            self.delete_object(bucket, key)
            yield key


    def delete_object(self, bucket, key):
        self.__client.delete_object(Bucket=bucket, Key=key)


    def exists(self, bucket, key):
        try:
            self.head(bucket, key)
            return True

        except ClientError as ex:
            if ex.response['Error']['Code'] == "404":
                return False

            raise


    def head(self, bucket, key):
        response = self.__client.head_object(Bucket=bucket, Key=key)
        return Head.construct(key, response)


    # ----------------------------------------------------------------------------------------------------------------

    def __retrieve_objects(self, bucket, prefix, next_token):
        if prefix:
            if next_token:
                response = self.__client.list_objects_v2(
                    Bucket=bucket,
                    Prefix=prefix,
                    ContinuationToken=next_token,
                    Delimiter=",",
                )
            else:
                response = self.__client.list_objects_v2(
                    Bucket=bucket,
                    Prefix=prefix,
                    Delimiter=",",
                )

        else:
            if next_token:
                response = self.__client.list_objects_v2(
                    Bucket=bucket,
                    ContinuationToken=next_token,
                    Delimiter=",",
                )
            else:
                response = self.__client.list_objects_v2(
                    Bucket=bucket,
                    Delimiter=",",
                )

        return response


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "S3Manager:{client:%s, resource_client:%s}" % (self.__client, self.__resource_client)


# --------------------------------------------------------------------------------------------------------------------

class S3PersistenceManager(PersistenceManager):
    """
    classdocs
    """

    __BUCKET = 'scs-persistence'

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def __key(dirname, filename):
        return '/'.join((dirname, filename))


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client, resource_client):
        """
        Constructor
        """
        self.__manager = S3Manager(client, resource_client)


    # ----------------------------------------------------------------------------------------------------------------

    def list(self, container, dirname):
        prefix_len = len(dirname) + 1
        objects = self.__manager.list_objects(container, 1, True, prefix=dirname)

        return tuple(obj.key[prefix_len:] for obj in objects)


    def exists(self, dirname, filename):
        key = self.__key(dirname, filename)

        return self.__manager.exists(self.__BUCKET, key)


    def load(self, dirname, filename, encryption_key=None):
        key = self.__key(dirname, filename)

        text, last_modified = self.__manager.retrieve_from_bucket(self.__BUCKET, key)

        if encryption_key:
            from scs_core.data.crypt import Crypt               # late import
            jstr = Crypt.decrypt(encryption_key, text)
        else:
            jstr = text

        return jstr, last_modified


    def save(self, text, dirname, filename, encryption_key=None):
        key = self.__key(dirname, filename)

        if encryption_key:
            from scs_core.data.crypt import Crypt               # late import
            saved_text = Crypt.encrypt(encryption_key, text)
        else:
            saved_text = text + '\n'

        self.__manager.put_object(saved_text, self.__BUCKET, key)


    def remove(self, dirname, filename):
        key = self.__key(dirname, filename)

        self.__manager.delete_object(self.__BUCKET, key)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def scs_path(cls):
        return cls.__BUCKET


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "S3PersistenceManager:{manager:%s}" % self.__manager


# --------------------------------------------------------------------------------------------------------------------

class Bucket(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, response):
        name = response.get('Name')
        creation_date = LocalizedDatetime(response.get('CreationDate'))

        return cls(name, creation_date)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, name, creation_date):
        """
        Constructor
        """
        self.__name = name                                  # string
        self.__creation_date = creation_date                # LocalizedDatetime


    def __lt__(self, other):
        return self.name < other.name


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['name'] = self.name
        jdict['creation-date'] = self.creation_date.as_iso8601()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def name(self):
        return self.__name


    @property
    def creation_date(self):
        return self.__creation_date


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Bucket:{name:%s, creation_date:%s}" % (self.name, self.creation_date)


# --------------------------------------------------------------------------------------------------------------------

class Head(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, key, response):
        last_modified = LocalizedDatetime(response.get('LastModified'))
        e_tag = response.get('ETag')
        size = response.get('ContentLength')
        content_type = response.get('ContentType')

        return cls(key, last_modified, e_tag, size, content_type)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, key, last_modified, e_tag, size, content_type):
        """
        Constructor
        """
        self.__key = key                                    # string
        self.__last_modified = last_modified                # LocalizedDatetime
        self.__e_tag = e_tag                                # string
        self.__size = int(size)                             # int
        self.__content_type = content_type                  # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['key'] = self.key
        jdict['last-modified'] = self.last_modified.as_iso8601()
        jdict['e-tag'] = self.e_tag
        jdict['size'] = self.size
        jdict['storage-class'] = self.content_type

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def key(self):
        return self.__key


    @property
    def last_modified(self):
        return self.__last_modified


    @property
    def e_tag(self):
        return self.__e_tag


    @property
    def size(self):
        return self.__size


    @property
    def content_type(self):
        return self.__content_type


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Head:{key:%s, last_modified:%s, e_tag:%s, size:%s, content_type:%s}" % \
               (self.key, self.last_modified, self.e_tag, self.size, self.content_type)


# --------------------------------------------------------------------------------------------------------------------

class Object(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, response):
        key = response.get('Key')
        last_modified = LocalizedDatetime(response.get('LastModified'))
        e_tag = response.get('ETag')
        size = response.get('Size')
        storage_class = response.get('StorageClass')

        return cls(key, last_modified, e_tag, size, storage_class)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, key, last_modified, e_tag, size, storage_class):
        """
        Constructor
        """
        self.__key = key                                    # string
        self.__last_modified = last_modified                # LocalizedDatetime
        self.__e_tag = e_tag                                # string
        self.__size = int(size)                             # int
        self.__storage_class = storage_class                # string


    def __lt__(self, other):
        if self.key < other.key:
            return True

        return self.e_tag < other.e_tag


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['key'] = self.key
        jdict['last-modified'] = self.last_modified.as_iso8601()
        jdict['e-tag'] = self.e_tag
        jdict['size'] = self.size
        jdict['storage-class'] = self.storage_class

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def key(self):
        return self.__key


    @property
    def last_modified(self):
        return self.__last_modified


    @property
    def e_tag(self):
        return self.__e_tag


    @property
    def size(self):
        return self.__size


    @property
    def storage_class(self):
        return self.__storage_class


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Object:{key:%s, last_modified:%s, e_tag:%s, size:%s, storage_class:%s}" % \
               (self.key, self.last_modified, self.e_tag, self.size, self.storage_class)


# --------------------------------------------------------------------------------------------------------------------

class Summary(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def none(cls):
        return cls(None, 0, None, 0)


    @classmethod
    def new(cls, path, obj: Object):
        return cls(path, 1, obj.last_modified, obj.size)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, path, objects, last_modified, size):
        """
        Constructor
        """
        self.__path = path                                  # string
        self.__objects = int(objects)                       # int
        self.__last_modified = last_modified                # LocalizedDatetime
        self.__size = int(size)                             # int


    def __lt__(self, other):
        return self.path < other.path


    # ----------------------------------------------------------------------------------------------------------------

    def is_none(self):
        return self.path is None


    def add(self, obj: Object):
        self.__objects += 1

        if self.last_modified is None or obj.last_modified > self.last_modified:
            self.__last_modified = obj.last_modified

        self.__size += obj.size


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['path'] = self.path
        jdict['objects'] = self.objects
        jdict['last-modified'] = self.last_modified.as_iso8601()
        jdict['size'] = self.size

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def path(self):
        return self.__path


    @property
    def objects(self):
        return self.__objects


    @property
    def last_modified(self):
        return self.__last_modified


    @property
    def size(self):
        return self.__size


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Summary:{path:%s, objects:%s, last_modified:%s, size:%s}" %  \
               (self.path, self.objects, self.last_modified, self.size)



"""
Created on 22 Nov 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://docs.aws.amazon.com/cognito/latest/developerguide/signing-up-users-in-your-app.html
https://docs.aws.amazon.com/cognito/latest/developerguide/user-pool-settings-policies.html

example document:
{"username": "506cd055-1978-4984-9f17-2fad77797fa1", "email": "bruno.beloff@southcoastscience.com",
"given-name": "Bruno", "family-name": "Beloff", "confirmation-status": "CONFIRMED", "enabled": true,
"email-verified": true, "is-super": true, "is-tester": true, "is-financial": true,
"created": "2023-04-20T11:45:21Z", "last-updated": "2023-06-26T14:39:17Z"}
"""

import re

from collections import OrderedDict

from scs_core.aws.data.dataset import DatasetItem

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserCredentials(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, email, password):
        """
        Constructor
        """
        self.__email = email                                        # string (email)
        self.__password = password                                  # string (AWS password)


    # ----------------------------------------------------------------------------------------------------------------

    def ok(self):
        if not self.email or not self.password:
            return False

        if not Datum.is_email_address(self.email):
            return False

        if not CognitoUserIdentity.is_valid_password(self.password):
            return False

        return True


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def email(self):
        return self.__email


    @email.setter
    def email(self, email):
        self.__email = email


    @property
    def password(self):
        return self.__password


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CognitoUserCredentials:{email:%s, password:%s}" %  (self.email, self.password)


# --------------------------------------------------------------------------------------------------------------------

class CognitoUserIdentity(DatasetItem, JSONable):
    """
    classdocs
    """

    __STATUSES = {
        'U': 'UNCONFIRMED',
        'C': 'CONFIRMED',
        'P': 'PASSWORD_RESET_REQUIRED',
        'F': 'FORCE_CHANGE_PASSWORD',
        'D': 'DISABLED'
    }

    @classmethod
    def status_codes(cls):
        return cls.__STATUSES.keys()


    @classmethod
    def status(cls, code):
        return cls.__STATUSES[code]


    __EMAIL_FUNCTIONS = {
        'UNCONFIRMED': 'RESEND_CONFIRMATION',
        'CONFIRMED': 'REQUEST_PASSWORD_RESET',
        'PASSWORD_RESET_REQUIRED': 'REQUEST_PASSWORD_RESET',
        'FORCE_CHANGE_PASSWORD': 'RESEND_TEMPORARY_PASSWORD',
        'DISABLED': None
    }


    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def is_valid_password(password):
        if not isinstance(password, str):
            return False

        if not 7 < len(password) < 99:
            return False

        if not re.findall(r'[0-9]', password):
            return False

        if not re.findall(r'[A-Z]', password):
            return False

        if not re.findall(r'[a-z]', password):
            return False

        if not re.findall(r'[\^$*.\[\]{}()?"!@#%&/\\,><\':;|_~`]', password):
            return False

        return True


    @staticmethod
    def ext_name(name):
        return '-' if name is None or name.strip() == '' else name


    @staticmethod
    def int_name(name):
        return None if name == '-' else name


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, None, None, None, None, None, None, None, None, None, None, None, None) if skeleton \
                else None

        username = jdict.get('username')
        email = jdict.get('email')
        password = jdict.get('password')
        given_name = cls.int_name(jdict.get('given-name'))
        family_name = cls.int_name(jdict.get('family-name'))

        confirmation_status = jdict.get('confirmation-status')
        enabled = jdict.get('enabled')
        email_verified = jdict.get('email-verified')

        is_super = jdict.get('is-super', False)
        is_tester = jdict.get('is-tester', False)
        is_financial = jdict.get('is-financial', False)

        created = LocalizedDatetime.construct_from_iso8601(jdict.get('created'))
        last_updated = LocalizedDatetime.construct_from_iso8601(jdict.get('last-updated'))

        return cls(username, created, confirmation_status, enabled, email_verified, email,
                   given_name, family_name, password, is_super, is_tester, is_financial, last_updated)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, username, created, confirmation_status, enabled, email_verified, email,
                 given_name, family_name, password, is_super, is_tester, is_financial, last_updated):
        """
        Constructor
        """

        self._username = username                                   # string email address or hash
        self._created = created                                     # LocalisedDatetime
        self.__confirmation_status = confirmation_status            # string
        self.__enabled = Datum.bool(enabled)                        # bool or None

        self.__email_verified = bool(email_verified)                # bool
        self.__email = email                                        # string
        self.__given_name = given_name                              # string
        self.__family_name = family_name                            # string
        self.__password = password                                  # string
        self.__is_super = bool(is_super)                            # bool
        self.__is_tester = bool(is_tester)                          # bool
        self.__is_financial = bool(is_financial)                    # bool

        self._last_updated = last_updated                           # LocalizedDatetime


    def __eq__(self, other):
        try:
            return self.username == other.username and self.confirmation_status == other.confirmation_status \
                   and self.enabled == other.enabled and self.email_verified == other.email_verified \
                   and self.email == other.email and self.given_name == other.given_name \
                   and self.family_name == other.family_name and self.is_super == other.is_super \
                   and self.is_tester == other.is_tester

        except (TypeError, AttributeError):
            return False


    def __lt__(self, other):
        # family_name...
        self_name = self.ext_name(self.family_name).lower()
        other_name = self.ext_name(other.family_name).lower()

        if self_name < other_name:
            return True

        if self_name > other_name:
            return False

        # given_name...
        self_name = self.ext_name(self.given_name).lower()
        other_name = self.ext_name(other.given_name).lower()

        if self_name < other_name:
            return True

        if self_name > other_name:
            return False

        # email...
        if self.email.lower() < other.email.lower():
            return True

        if self.email.lower() > other.email.lower():
            return False

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def references(self, item):
        return item.username == self.username


    def copy_pk(self, other):
        pass


    def save(self, db_user):
        raise NotImplementedError


    def delete(self, db_user):
        raise NotImplementedError


    @property
    def index(self):
        return self.email


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def email_function(self):
        try:
            return self.__EMAIL_FUNCTIONS[self.confirmation_status]

        except KeyError:
            raise ValueError(self.confirmation_status)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.username is not None:
            jdict['username'] = self.username

        jdict['email'] = self.email

        if self.password is not None:
            jdict['password'] = self.password

        jdict['given-name'] = self.ext_name(self.given_name)
        jdict['family-name'] = self.ext_name(self.family_name)

        if self.confirmation_status is not None:
            jdict['confirmation-status'] = self.confirmation_status

        if self.enabled is not None:
            jdict['enabled'] = self.enabled

        jdict['email-verified'] = self.email_verified

        if self.is_super:
            jdict['is-super'] = self.is_super

        if self.is_tester:
            jdict['is-tester'] = self.is_tester

        if self.is_financial:
            jdict['is-financial'] = self.is_financial

        if self.created is not None:
            jdict['created'] = self.created.as_iso8601()

        if self.last_updated is not None:
            jdict['last-updated'] = self.last_updated.as_iso8601()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def username(self):
        return self._username


    @property
    def created(self):
        return self._created


    @property
    def confirmation_status(self):
        return self.__confirmation_status


    @property
    def enabled(self):
        return self.__enabled


    @property
    def email_verified(self):
        return self.__email_verified


    @property
    def email(self):
        return self.__email


    @property
    def given_name(self):
        return self.__given_name


    @property
    def family_name(self):
        return self.__family_name


    @property
    def password(self):
        return self.__password


    @property
    def is_super(self):
        return self.__is_super


    @property
    def is_tester(self):
        return self.__is_tester


    @property
    def is_financial(self):
        return self.__is_financial


    @property
    def last_updated(self):
        return self._last_updated


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{username:%s, created:%s, confirmation_status:%s, enabled:%s, " \
               "email_verified:%s, email:%s, given_name:%s, family_name:%s, is_super:%s, is_tester:%s, " \
               "is_financial:%s, last_updated:%s}" % \
               (self.username, self.created, self.confirmation_status, self.enabled,
                self.email_verified, self.email, self.given_name, self.family_name, self.is_super, self.is_tester,
                self.is_financial, self.last_updated)



"""
Created on 9 Dec 2020

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

The A4CalibratedDatum is a normalised electrochem output voltage. Its vCal field represents the calibrated weV
minus the calibrated aeV.

(Obsolete: The vXCal field should only be computed when the NO2 concentration is known precisely, so this should be
done in the preprocessing phase for the O3 model.)

example document:
{"weV": 0.39519, "aeV": 0.39963, "weC": 0.00627, "cnc": 61.9, "vCal": -9.801, "xCal": -12.601}
"""

from collections import OrderedDict

from scs_core.data.datum import Datum

from scs_core.gas.a4.a4_calib import A4Calib
from scs_core.gas.a4.a4_datum import A4Datum


# --------------------------------------------------------------------------------------------------------------------

class A4Calibrator(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, calib: A4Calib):
        """
        Constructor
        """
        self.__we_elc_v = round(calib.we_elc_mv / 1000.0, 3)                        # we_electronic_zero_mv
        self.__ae_elc_v = round(calib.ae_elc_mv / 1000.0, 3)                        # ae_electronic_zero_mv

        self.__we_sens_v = round(calib.we_sens_mv / 1000.0, 6)                      # we_sensitivity_mv_ppb

        if calib.is_ox_sensor():
            self.__we_no2_x_sens_v = round(calib.we_no2_x_sens_mv / 1000.0, 6)      # we_cross_sensitivity_no2_mv_ppb
        else:
            self.__we_no2_x_sens_v = None


    # ----------------------------------------------------------------------------------------------------------------

    def calibrate(self, datum):
        v_cal = self.__calibrate(datum, self.we_sens_v)
        x_cal = self.__calibrate(datum, self.we_no2_x_sens_v)

        return A4CalibratedDatum(datum.we_v, datum.ae_v, datum.we_c, datum.cnc, v_cal, x_cal=x_cal)


    def __calibrate(self, datum, we_sens_v):
        if we_sens_v is None:
            return None

        # zero offset...
        we_v_zero_cal = datum.we_v - self.we_elc_v
        ae_v_zero_cal = datum.ae_v - self.ae_elc_v

        # remove noise...
        v_zero_cal = we_v_zero_cal - ae_v_zero_cal

        # gain...
        return v_zero_cal / we_sens_v


    def set_we_v_zero_x_cal(self, datum, no2_cnc):
        # we_v_x = no2_cnc * self.we_no2_x_sens_v
        # datum.we_v_zero_x_cal = we_v_x - self.we_elc_v
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_elc_v(self):
        return self.__we_elc_v


    @property
    def ae_elc_v(self):
        return self.__ae_elc_v


    @property
    def we_sens_v(self):
        return self.__we_sens_v


    @property
    def we_no2_x_sens_v(self):
        return self.__we_no2_x_sens_v


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4Calibrator:{we_elc_v:%s, ae_elc_v:%s, we_sens_v:%s, we_no2_x_sens_v:%s}" % \
               (self.we_elc_v, self.ae_elc_v, self.we_sens_v, self.we_no2_x_sens_v)


# --------------------------------------------------------------------------------------------------------------------

class A4CalibratedDatum(A4Datum):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        we_v = jdict.get('weV')
        ae_v = jdict.get('aeV')
        we_c = jdict.get('weC')
        cnc = jdict.get('cnc')

        v_cal = jdict.get('vCal')
        x_cal = jdict.get('xCal')

        return cls(we_v, ae_v, we_c, cnc, v_cal, x_cal)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, we_v, ae_v, we_c, cnc, v_cal, x_cal=None):
        """
        Constructor
        """
        super().__init__(we_v, ae_v, we_c, cnc)

        self.__v_cal = Datum.float(v_cal, 3)                        # calibrated ppb
        self.__x_cal = Datum.float(x_cal, 3)                        # calibrated ppb according to cross-sensitivity


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['weV'] = self.we_v
        jdict['aeV'] = self.ae_v
        jdict['weC'] = self.we_c                                # may be None
        jdict['cnc'] = self.cnc                                 # may be None

        if self.v_cal is not None:
            jdict['vCal'] = self.v_cal

        if self.x_cal is not None:
            jdict['xCal'] = self.x_cal

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def v_cal(self):
        return self.__v_cal


    @property
    def x_cal(self):
        return self.__x_cal


    # @x_cal.setter
    # def x_cal(self, x_cal):
    #     self.__x_cal = round(x_cal, 9)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "A4CalibratedDatum:{we_v:%s, ae_v:%s, we_c:%s, cnc:%s, v_cal:%s, x_cal:%s}" % \
               (self.we_v, self.ae_v, self.we_c, self.cnc, self.v_cal, self.x_cal)



"""
Created on 21 Jan 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from abc import ABC, abstractmethod
from collections import OrderedDict

from scs_core.data.str import Str
from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class DatasetItem(ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @abstractmethod
    def references(self, other):
        pass


    @abstractmethod
    def copy_pk(self, other):
        pass


    @abstractmethod
    def save(self, db_username):
        pass


    @abstractmethod
    def delete(self, db_username):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @property
    @abstractmethod
    def index(self):                                        # a scalar that is unique within the dataset
        return None


    @property
    @abstractmethod
    def last_updated(self):                                 # LocalizedDatetime
        return None


# --------------------------------------------------------------------------------------------------------------------

class Dataset(object):
    """
    classdocs
    """
    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, db_username, latest_import, simulate=False):
        """
        Constructor
        """
        self.__db_username = db_username                    # string
        self.__latest_import = latest_import                # LocalizedDatetime
        self.__simulate = simulate                          # bool

        self.__items = OrderedDict()                        # dict of index: Indexable
        self.__logger = Logging.getLogger()


    def __len__(self):
        return len(self.__items)


    def __contains__(self, item):
        for dataset_item in self.__items.values():
            # print("item: %s" % item)
            # print("dataset_item: %s" % dataset_item)
            if dataset_item.references(item):
                return True

        return False


    # ----------------------------------------------------------------------------------------------------------------

    def clear(self):
        self.__items = OrderedDict()


    def add(self, item: DatasetItem):
        self.__items[item.index] = item


    def extend(self, items):
        for item in items:
            self.add(item)


    def update_with(self, item: DatasetItem, insert_only=False):
        try:
            retrieved_item = self.__items[item.index]

            if insert_only:
                return

            # no changes...
            if item == retrieved_item:
                return

            # AWS item is newer - ignore old-world item...
            if self.latest_import and retrieved_item.last_updated > self.latest_import:
                self.__logger.warn('WARNING: old-world item ignored: %s' % item)
                return

            # old-world item is newer...
            item.copy_pk(retrieved_item)

            if not self.simulate:
                item.save(self.db_username)

            self.__items[item.index] = item
            self.__logger.warn('updated: %s' % item)

        except KeyError:
            # no AWS item...
            if not self.simulate:
                item.save(self.db_username)

            self.__items[item.index] = item
            self.__logger.warn('inserted: %s' % item)


    def delete_unreferenced(self, references):
        for index, item in list(self.__items.items()):
            if item in references:
                continue

            if not self.simulate:
                item.delete(self.db_username)

            del self.__items[index]
            self.__logger.warn("deleted: %s" % item)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def db_username(self):
        return self.__db_username


    @property
    def latest_import(self):
        return self.__latest_import


    @property
    def simulate(self):
        return self.__simulate


    def item(self, index):
        return self.__items[index]                          # may raise KeyError


    def keys(self):
        return self.__items.keys()


    def values(self):
        return self.__items.values()


    def items(self):
        return self.__items.items()


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "Dataset:{db_username:%s, latest_import:%s, simulate:%s, items:%s}" % \
               (self.db_username, self.latest_import, self.simulate, Str.collection(self.__items))



"""
Created on 07 Apr 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
"""

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint
from scs_core.aws.manager.configuration.configuration_intercourse import ConfigurationRequest, ConfigurationResponse


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('ConfAPI/ConfigurationFinder',
                   'https://p18hyi3w56.execute-api.us-west-2.amazonaws.com/default/ConfigurationFinder')


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationFinder(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        super().__init__(reporter=reporter)


    # ----------------------------------------------------------------------------------------------------------------

    def find(self, token, tag_filter, exact_match, response_mode):
        if self._reporter:
            self._reporter.reset()

        request = ConfigurationRequest(tag_filter, exact_match, response_mode)

        for item in self._get_blocks(Endpoint.url(), token, ConfigurationResponse, params=request.params()):
            yield item



"""
Created on 30 Dec 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Global positioning system fix data
$xxGGA,time,lat,NS,long,EW,quality,numSV,HDOP,alt,M,sep,M,diffAge,diffStation*cs

example sentence:
$GPGGA,092725.00,4717.11399,N,00833.91590,E,1,08,1.01,499.6,M,48.0,M,,*5B

example values:
GPGGA:{time:GPTime:{time:141058.00}, loc:GPLoc:{lat:5049.38432, ns:N, lng:00007.37801, ew:W}, quality:2, num_sv:06,
hdop:3.10, alt:37.5, sep:45.4, diff_age:None, diff_station:0000}

GPGGA:{time:GPTime:{time:140047.00}, loc:GPLoc:{lat:None, ns:None, lng:None, ew:None}, quality:0, num_sv:00,
hdop:99.99, alt:None, sep:None, diff_age:None, diff_station:None}

https://www.nmea.org
https://en.wikipedia.org/wiki/NMEA_0183
"""

from scs_core.position.nmea.gploc import GPLoc
from scs_core.position.nmea.gptime import GPTime
from scs_core.position.nmea.nmea_sentence import NMEASentence


# --------------------------------------------------------------------------------------------------------------------

class GPGGA(NMEASentence):
    """
    classdocs
    """

    MESSAGE_IDS = ("$GNGGA", "$GPGGA")

    QUALITY_NO_FIX =                0
    QUALITY_AUTONOMOUS_GNSS =       1
    QUALITY_DIFFERENTIAL_GNSS =     2
    QUALITY_ESTIMATED_FIX =         6


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct(cls, r):
        if r.message_id not in cls.MESSAGE_IDS:
            raise TypeError("invalid sentence:%s" % r)

        time = GPTime(r.str(1))

        lat = r.str(2)
        ns = r.str(3)

        lng = r.str(4)
        ew = r.str(5)

        loc = GPLoc(lat, ns, lng, ew)

        quality = r.int(6)
        num_sv = r.int(7)
        hdop = r.float(8, 3)
        alt = r.float(9, 2)
        sep = r.float(11, 2)

        diff_age = r.float(13, 3)
        diff_station = r.str(14)

        return GPGGA(r.message_id, time, loc, quality, num_sv, hdop, alt, sep, diff_age, diff_station)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, message_id, time, loc, quality, num_sv, hdop, alt, sep, diff_age, diff_station):
        """
        Constructor
        """
        super().__init__(message_id)

        self.__time = time                          # GPTime
        self.__loc = loc                            # GPLoc

        self.__quality = quality                    # int
        self.__num_sv = num_sv                      # int
        self.__hdop = hdop                          # float(2)
        self.__alt = alt                            # float(1) - altitude (metres)
        self.__sep = sep                            # float(1) - geoid separation (metres)

        self.__diff_age = diff_age                  # float(3) - age of differential corrections (seconds)
        self.__diff_station = diff_station          # string - ID of station providing differential corrections


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def time(self):
        return self.__time


    @property
    def loc(self):
        return self.__loc


    @property
    def quality(self):
        return self.__quality


    @property
    def num_sv(self):
        return self.__num_sv


    @property
    def hdop(self):
        return self.__hdop


    @property
    def alt(self):
        return self.__alt


    @property
    def sep(self):
        return self.__sep


    @property
    def diff_age(self):
        return self.__diff_age


    @property
    def diff_station(self):
        return self.__diff_station


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GPGGA:{source:%s, time:%s, loc:%s, quality:%s, num_sv:%s, hdop:%s, alt:%s, sep:%s, " \
               "diff_age:%s, diff_station:%s}" % \
                    (self.source, self.time, self.loc, self.quality, self.num_sv, self.hdop, self.alt, self.sep,
                     self.diff_age, self.diff_station)



"""
Created on 25 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import struct


# --------------------------------------------------------------------------------------------------------------------

class EEPROMImage(object):
    """
    Load a .eep file
    """
    @classmethod
    def construct_from_file(cls, filename, min_size=0):
        content = []

        # file content...
        f = open(filename, "rb")

        try:
            while True:
                byte = f.read(1)

                if not byte:
                    break

                content.append(struct.unpack('B', byte)[0])

        except RuntimeError:
            return None

        finally:
            f.close()

        # padding...
        if min_size and len(content) < min_size:
            content += [0] * (min_size - len(content))

        return EEPROMImage(content)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, content):
        """
        Constructor
        """
        self.__content = content


    def __len__(self):
        return len(self.__content)


    def __eq__(self, other):
        try:
            return self.content == other.content

        except (TypeError, AttributeError):
            return False


    def __ne__(self, other):
        return not self.__eq__(other)


    # ----------------------------------------------------------------------------------------------------------------

    def formatted(self, width):
        addr = 0

        while addr < len(self.__content):
            values = self.__content[addr: addr + width]

            hexs = ' '.join("%02x" % value for value in values)
            hexs = hexs.ljust(width * 3)

            chrs = ''.join(chr(value) if 31 < value < 128 else '.' for value in values)    # non-printable chars are '.'

            print("0x%04x: %s        %s" % (addr, hexs, chrs))

            addr += width


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def content(self):
        return self.__content


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "EEPROMImage:{content:%s}" % self.content



"""
Created on 04 Dec 2020

@author: Jade Page (jade.page@southcoastscience.com)

"""

import sys


# --------------------------------------------------------------------------------------------------------------------

class SagemakerTrialManager(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, client):
        """
        Constructor
        """
        self.__client = client


    # ----------------------------------------------------------------------------------------------------------------

    def list_experiments(self, time_order):
        names = []
        next_token = None

        try:
            while True:
                res, next_token = self.retrieve_experiments(time_order, next_token)
                experiments = res.get("ExperimentSummaries")

                for item in experiments:
                    name = item.get("ExperimentName")
                    names.append(name)

                if not next_token:
                    return names

        except self.__client.exceptions.ResourceNotFound:
            print("Resource not found - ensure your auth is for the correct account", file=sys.stderr)


    def retrieve_experiments(self, time_order, next_token):
        next_token2 = None

        if next_token:
            response = self.__client.list_experiments(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
            )
        else:
            response = self.__client.list_experiments(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    def list_components(self, experiment_string, trial_string, time_order):
        components = []
        next_token = None

        try:
            if trial_string:
                while True:
                    res, next_token = self.retrieve_trial_components(trial_string, time_order, next_token)
                    components.append(res)

                    if not next_token:
                        return components

            elif experiment_string:
                while True:
                    res, next_token = self.retrieve_experiment_components(experiment_string, time_order, next_token)
                    components.append(res)

                    if not next_token:
                        return components

        except self.__client.exceptions.ResourceNotFound:
            print("Resource not found - please check the filter string", file=sys.stderr)


    def retrieve_trial_components(self, trial_string, time_order, next_token):
        next_token2 = None

        if next_token:
            response = self.__client.list_trial_components(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                TrialName=trial_string,
                NextToken=next_token
            )
        else:
            response = self.__client.list_trial_components(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                TrialName=trial_string
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    def retrieve_experiment_components(self, experiment_string, time_order, next_token):
        next_token2 = None

        if next_token:
            response = self.__client.list_trial_components(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                ExperimentName=experiment_string,
                NextToken=next_token
            )
        else:
            response = self.__client.list_trial_components(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                ExperimentName=experiment_string
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    def list_trial_names_filter(self, experiment_string, time_order):
        names = []
        next_token = None

        try:
            while True:
                res, next_token = self.retrieve_filtered_trials(experiment_string, time_order, next_token)
                models = res.get("TrialSummaries")

                for item in models:
                    name = item.get("TrialName")
                    names.append(name)

                if not next_token:
                    return names

        except self.__client.exceptions.ResourceNotFound:
            print("Resource not found - please check the filter string", file=sys.stderr)


    def list_trial_names(self, time_order):
        names = []
        next_token = None

        try:
            while True:
                res, next_token = self.retrieve_trials(time_order, next_token)
                models = res.get("TrialSummaries")

                for item in models:
                    name = item.get("TrialName")
                    names.append(name)

                if not next_token:
                    return names

        except self.__client.exceptions.ResourceNotFound:
            print("Resource not found - ensure your auth is for the correct account", file=sys.stderr)


    def retrieve_trials(self, time_order, next_token=None):
        next_token2 = None

        if next_token:
            response = self.__client.list_trials(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                NextToken=next_token
            )
        else:
            response = self.__client.list_trials(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    def retrieve_filtered_trials(self, experiment_string, time_order, next_token=None):
        next_token2 = None

        if next_token:
            response = self.__client.list_trials(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                ExperimentName=experiment_string,
                NextToken=next_token
            )
        else:
            response = self.__client.list_trials(
                SortBy='CreationTime' if time_order else 'Name',
                SortOrder='Descending' if time_order else 'Ascending',
                MaxResults=100,
                ExperimentName=experiment_string
            )

        if "NextToken" in response:
            next_token2 = response.get("NextToken")

        return response, next_token2


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "SagemakerTrialManager:{client:%s}" % self.__client



"""
Created on 24 Feb 2021

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

JSON example:
{"pulled-on": "2021-02-27T08:37:09Z", "success": true,
"installed": ["scs_core", "scs_dev", "scs_dfe_eng", "scs_host_cpc", "scs_mfr", "scs_psu"],
"pulled": ["scs_core", "scs_dev", "scs_dfe_eng", "scs_host_cpc", "scs_mfr", "scs_psu"],
"excluded": []}
"""

from collections import OrderedDict
from subprocess import Popen, PIPE, TimeoutExpired

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import PersistentJSONable

from scs_core.sys.filesystem import Filesystem


# --------------------------------------------------------------------------------------------------------------------

class GitPull(PersistentJSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    __DIR_PREFIX = 'scs_'
    __EXCLUSIONS = ('scs_exegesis', 'scs_exegesis_modelling', 'scs_experimental', 'scs_inference', 'scs-installer')

    @classmethod
    def excludes(cls, name):
        return name in cls.__EXCLUSIONS


    @classmethod
    def dirs(cls, root):
        items = Filesystem.ls(root)

        if not items:
            return tuple()

        return tuple(item.name for item in items if item.is_directory and item.name.startswith(cls.__DIR_PREFIX))


    @classmethod
    def is_clone(cls, path):
        items = Filesystem.ls(path)

        if not items:
            return False

        return '.git' in [item.name for item in items]


    @classmethod
    def pull_repo(cls, path, timeout):
        try:
            p = Popen(['git', '-C', path, 'pull'], stdout=PIPE, stderr=PIPE)
            stdout_bytes, stderr_bytes = p.communicate(timeout=timeout)

            success = p.returncode == 0
            stdout = stdout_bytes.decode()
            stderr = stderr_bytes.decode()

            return success, stdout, stderr

        except TimeoutExpired:
            raise TimeoutError(timeout)


    # ----------------------------------------------------------------------------------------------------------------

    __FILENAME = "git_pull.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return cls(None, False, [], [], []) if skeleton else None

        pulled_on = LocalizedDatetime.construct_from_jdict(jdict.get('pulled-on'))
        success = jdict.get('success')

        installed = jdict.get('installed')
        pulled = jdict.get('pulled')
        excluded = jdict.get('excluded')

        return cls(pulled_on, success, installed, pulled, excluded)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, pulled_on, success, installed, pulled, excluded):
        """
        Constructor
        """
        super().__init__()

        self.__pulled_on = pulled_on                    # LocalizedDatetime
        self.__success = success                        # bool

        self.__installed = installed                    # array of strings
        self.__pulled = pulled                          # array of strings
        self.__excluded = excluded                      # array of strings


    def __eq__(self, other):
        try:
            return self.pulled_on == other.pulled_on and self.success == other.success and \
                   self.installed == other.installed and self.pulled == other.pulled and \
                   self.excluded == other.excluded

        except (TypeError, AttributeError):
            return False


    # ----------------------------------------------------------------------------------------------------------------

    def is_comprehensive(self):
        return self.installed == self.pulled


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['pulled-on'] = None if self.pulled_on is None else self.pulled_on.as_iso8601()
        jdict['success'] = self.success

        jdict['installed'] = self.installed
        jdict['pulled'] = self.pulled
        jdict['excluded'] = self.excluded

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def pulled_on(self):
        return self.__pulled_on


    @property
    def success(self):
        return self.__success


    @property
    def installed(self):
        return self.__installed


    @property
    def pulled(self):
        return self.__pulled


    @property
    def excluded(self):
        return self.__excluded


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "GitPull:{pulled_on:%s, success:%s, installed:%s, pulled:%s, excluded:%s}" % \
               (self.pulled_on, self.success, self.installed, self.pulled, self.excluded)



"""
Created on 14 Jun 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

document example:
{"scs-be2-3": {"device-tag": "scs-be2-3", "recipients": [{"email": "bbeloff@me.com", "json-message": true},
{"email": "bruno.beloff@southcoastscience.com", "json-message": false}], "suspended": false}}
"""

from collections import OrderedDict

from scs_core.data.json import JSONable, PersistentJSONable
from scs_core.data.str import Str

from scs_core.email.email import EmailRecipient


# --------------------------------------------------------------------------------------------------------------------

class DeviceMonitorSpecification(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if jdict is None:
            return None

        device_tag = jdict.get('device-tag')

        recipients = {}
        for recipient_jdict in jdict.get('recipients'):
            recipient = EmailRecipient.construct_from_jdict(recipient_jdict)
            recipients[recipient.email_address] = recipient

        is_suspended = jdict.get('suspended')

        return cls(device_tag, recipients, is_suspended)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_tag, recipients, is_suspended):
        """
        Constructor
        """
        super().__init__()

        self.__device_tag = device_tag                              # string
        self.__recipients = recipients                              # dict of email: DeviceMonitorRecipient
        self.__is_suspended = bool(is_suspended)                    # bool


    def __len__(self):
        return len(self.__recipients)


    # ----------------------------------------------------------------------------------------------------------------

    def add(self, recipient: EmailRecipient):
        self.__recipients[recipient.email_address] = recipient


    def discard(self, email_address):
        try:
            del self.__recipients[email_address]
        except KeyError:
            pass


    # ----------------------------------------------------------------------------------------------------------------

    def matches_tag(self, device_tag, exact):
        if device_tag is None:
            return True

        return (exact and device_tag == self.device_tag) or (not exact and device_tag in self.device_tag)


    def matches_email(self, email_address, exact):
        if email_address is None:
            return True

        for recipient in self.__recipients:
            if (exact and email_address == recipient) or (not exact and email_address in recipient):
                return True

        return False


    def contains_email(self, email_address):
        return email_address in self.__recipients


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['device-tag'] = self.device_tag
        jdict['recipients'] = self.recipients
        jdict['suspended'] = self.is_suspended

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def device_tag(self):
        return self.__device_tag


    @property
    def recipients(self):
        return sorted(self.__recipients.values())


    @property
    def is_suspended(self):
        return self.__is_suspended


    @is_suspended.setter
    def is_suspended(self, is_suspended):
        self.__is_suspended = is_suspended


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        recipients = Str.collection(self.recipients)

        return "DeviceMonitorSpecification:{device_tag:%s, recipients:%s, is_suspended:%s}" %  \
            (self.device_tag, recipients, self.is_suspended)


# --------------------------------------------------------------------------------------------------------------------

class DeviceMonitorSpecificationList(PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "device_monitor_specification_list.json"

    @classmethod
    def persistence_location(cls):
        return cls.aws_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if jdict is None:
            return None

        device_dict = {device_tag: DeviceMonitorSpecification.construct_from_jdict(specification_jdict)
                       for device_tag, specification_jdict in jdict.items()}

        return cls(device_dict)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, device_dict):
        """
        Constructor
        """
        super().__init__()

        self.__device_dict = device_dict                            # dict of device-tag: DeviceMonitorSpecification


    def __len__(self):
        return len(self.__device_dict)


    def __contains__(self, item):
        return item in self.__device_dict


    # ----------------------------------------------------------------------------------------------------------------

    def set_suspended(self, device_tag, is_suspended):
        self.__device_dict[device_tag].is_suspended = is_suspended              # may raise KeyError

        return self.__device_dict[device_tag]


    def add(self, device_tag, recipient: EmailRecipient):
        if device_tag not in self.__device_dict:
            self.__device_dict[device_tag] = DeviceMonitorSpecification(device_tag, {}, False)

        self.__device_dict[device_tag].add(recipient)

        return self.__device_dict[device_tag]


    def discard(self, device_tag, email_address):
        device_dict = {}

        # all devices...
        if device_tag is None:
            for device_tag in self.__device_dict:
                self.__device_dict[device_tag].discard(email_address)

                if len(self.__device_dict[device_tag]):
                    device_dict[device_tag] = self.__device_dict[device_tag]

        # specific device...
        elif device_tag not in self.__device_dict:
            device_dict = {}

        else:
            self.__device_dict[device_tag].discard(email_address)

            if len(self.__device_dict[device_tag]):
                device_dict = {device_tag: self.__device_dict[device_tag]}

        return DeviceMonitorSpecificationList(device_dict)


    def filter(self, email_address=None, device_tag=None, exact=False):
        device_dict = {}

        for specification in self.__device_dict.values():
            if not specification.matches_tag(device_tag, exact):
                continue

            if not specification.matches_email(email_address, exact):
                continue

            device_dict[specification.device_tag] = self.__device_dict[specification.device_tag]

        return DeviceMonitorSpecificationList(device_dict)


    def filter_devices(self, device_tags):
        device_dict = {key: self.__device_dict[key] for key in self.__device_dict.keys() & device_tags}

        return DeviceMonitorSpecificationList(device_dict)


    def subset(self, email_address):
        device_dict = {device_tag: specification for device_tag, specification in self.__device_dict.items()
                       if specification.contains_email(email_address)}

        return DeviceMonitorSpecificationList(device_dict)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.device_dict


    # ----------------------------------------------------------------------------------------------------------------

    def specification(self, device_tag):
        try:
            return self.__device_dict[device_tag]
        except KeyError:
            return None


    @property
    def device_dict(self):
        device_dict = {device_tag: specification for device_tag, specification in sorted(self.__device_dict.items())
                       if len(specification)}

        return device_dict


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "DeviceMonitorSpecificationList:{device_dict:%s}" %  Str.collection(self.device_dict)



"""
Created on 9 Jan 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

https://stackoverflow.com/questions/845058/how-to-get-the-line-count-of-a-large-file-cheaply-in-python
https://stackoverflow.com/questions/33626623/the-most-efficient-way-to-remove-first-n-elements-in-a-list
"""

from abc import ABC, abstractmethod
from collections import OrderedDict

from scs_core.csv.csv_reader import CSVReader
from scs_core.csv.csv_writer import CSVWriter

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.json import JSONable, JSONify, PersistentJSONable
from scs_core.data.str import Str

from scs_core.sys.filesystem import Filesystem


# --------------------------------------------------------------------------------------------------------------------

class LogEntry(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))
        event = jdict.get('event')

        return cls(rec, event)


    @classmethod
    def construct(cls, event):
        return cls(LocalizedDatetime.now(), event)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, rec, event):
        """
        Constructor
        """
        self.__rec = rec                                    # LocalizedDatetime
        self.__event = event                                # string


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['rec'] = self.rec.as_iso8601()
        jdict['event'] = self.event

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def rec(self):
        return self.__rec


    @property
    def event(self):
        return self.__event


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "LogEntry:{rec:%s, event:%s}" % (self.rec, self.event)


# --------------------------------------------------------------------------------------------------------------------

class CSVLog(PersistentJSONable, ABC):
    """
    classdocs
    """

    __TRIM_MARGIN =         0.2                             # 20%
    __LOG_DIR =             "log"                           # hard-coded rel path

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def log_dir(cls):
        return cls.__LOG_DIR


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    @abstractmethod
    def max_permitted_entries(cls):
        pass


    @classmethod
    @abstractmethod
    def construct_entry(cls, event):
        pass


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def rows(cls, manager):
        if not cls.exists(manager):
            return 0                                                # no log, so no errors

        dirname, filename = cls.persistence_location()
        abs_filename = manager.abs_filename(dirname, filename)

        num_lines = sum(1 for _ in open(abs_filename))

        return 0 if not num_lines else num_lines - 1                # ignore CSV header row


    @classmethod
    def trim(cls, manager, max_entries):
        if not cls.exists(manager):
            return

        dirname, filename = cls.persistence_location()
        abs_filename = manager.abs_filename(dirname, filename)

        with open(abs_filename) as f:
            lines = f.readlines()

        excess = len(lines) - max_entries

        if excess < 1:
            return

        remove = excess + int(max_entries * cls.__TRIM_MARGIN)

        del lines[1:remove]

        with open(abs_filename, "w") as f:
            f.writelines(lines)


    @classmethod
    def load(cls, manager, encryption_key=None, skeleton=False):
        if not cls.exists(manager):
            return cls.construct_from_jdict(None, skeleton=skeleton)

        dirname, filename = cls.persistence_location()
        text, last_modified = manager.load(dirname, filename)

        reader = CSVReader(text.splitlines())
        jdict = [cls.loads(row) for row in reader.rows()]

        try:
            obj = cls.construct_from_jdict(jdict, skeleton=skeleton)
            obj._last_modified = last_modified
            return obj

        except (AttributeError, TypeError):
            return None


    @classmethod
    def save_event(cls, manager, event, trim=False):
        cls.__save_entries(manager, [cls.construct_entry(event)])

        if trim:
            cls.trim(manager, cls.max_permitted_entries())


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __save_entries(cls, manager, entries):
        dirname, filename = cls.persistence_location()
        abs_filename = manager.abs_filename(dirname, filename)

        Filesystem.mkdir(manager.abs_dirname(dirname))
        append = cls.exists(manager)

        writer = None

        try:
            writer = CSVWriter(filename=abs_filename, append=append)

            for entry in entries:
                writer.write(JSONify.dumps(entry))

        finally:
            if writer:
                writer.close()


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, entries):
        """
        Constructor
        """
        super().__init__()

        self.__entries = entries                            # array of JSONable


    def __len__(self):
        return len(self.entries)


    # ----------------------------------------------------------------------------------------------------------------

    def save(self, manager, encryption_key=None):
        self.__save_entries(manager, self.entries)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        return self.__entries


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def entries(self):
        return self.__entries


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{entries:%s}" % Str.collection(self.entries)



"""
Created on 9 Feb 2022

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

* resend_confirmation - re-send confirmation email when account is created using the API
* resend_temporary_password - re-send temporary password that was created by the security_import system
* request_reset_password - re-send a code needed to do a password reset
* do_reset_password - perform a password reset using the code
* do_set_password - change password when in force reset mode created by security_import system

https://docs.python.org/3/howto/enum.html#planet

document example:
"EmailSent"
"""

import requests

from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.data.json import JSONify


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CogUsrPwdAPI/CognitoUserPassword',
                   'https://df46l72wl7.execute-api.us-west-2.amazonaws.com/default/CognitoUserPassword')


# --------------------------------------------------------------------------------------------------------------------

class CognitoPasswordManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def send_email(self, email):
        url = Endpoint.url('send-email')
        payload = {'email': email}

        response = requests.post(url, headers=self._auth_headers(), data=JSONify.dumps(payload))
        self._check_response(response)


    def do_reset_password(self, email, code, new_password):
        url = Endpoint.url('reset')
        payload = {'email': email, 'reset_code': code, 'new_password': new_password}

        response = requests.post(url, headers=self._auth_headers(), data=JSONify.dumps(payload))
        self._check_response(response)


    def do_set_password(self, email, new_password, session):
        url = Endpoint.url('respond')
        payload = {'email': email, 'new_password': new_password, 'session': session}

        response = requests.post(url, headers=self._auth_headers(), data=JSONify.dumps(payload))
        self._check_response(response)



"""
Created on 19 Sep 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

from collections import OrderedDict

from scs_core.data.datum import Datum
from scs_core.data.json import JSONable


# --------------------------------------------------------------------------------------------------------------------

class PIDDatum(JSONable):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        we_v = jdict.get('weV')

        we_c = jdict.get('weC')
        cnc = jdict.get('cnc')

        return cls(we_v, we_c, cnc)


    @classmethod
    def construct(cls, calib, baseline, tc, temp, we_v):
        if calib is None or tc is None:
            return PIDDatum(we_v)

        # weC...
        we_c = cls.__we_c(calib, tc, temp, we_v)

        if we_c is None:
            return PIDDatum(we_v)

        # cnc...
        cnc = cls.__cnc(calib, we_c)

        baselined_cnc = cnc + baseline.offset               # a positive offset causes the value to be raised

        return cls(we_v, we_c, baselined_cnc)


    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def __we_c(cls, calib, tc, temp, we_v):
        """
        Compute weC from sensor temperature compensation of weV
        """
        offset_v = calib.pid_elc_mv / 1000.0

        response_v = we_v - offset_v                    # remove electronic zero
        response_c = tc.correct(temp, response_v)       # correct the response component

        if response_c is None:
            return None

        we_c = response_c + offset_v                    # replace electronic zero

        return we_c


    @classmethod
    def __cnc(cls, calib, we_c):
        """
        Compute cnc from weC
        """
        if we_c is None:
            return None

        offset_v = calib.pid_elc_mv / 1000.0

        response_c = we_c - offset_v                    # remove electronic zero
        cnc = response_c / calib.pid_sens_v_ppb         # units are Volts / ppb

        corrected_cnc = cnc if calib.bump_sensitivity is None else cnc / calib.bump_sensitivity

        return corrected_cnc


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, we_v, we_c=None, cnc=None):
        """
        Constructor
        """
        self.__we_v = Datum.float(we_v, 5)              # uncorrected working electrode output      Volts

        self.__we_c = Datum.float(we_c, 5)              # corrected working electrode voltage       Volts
        self.__cnc = Datum.float(cnc, 1)                # gas concentration                         ppb


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        jdict['weV'] = self.we_v

        jdict['weC'] = self.we_c                        # may be None
        jdict['cnc'] = self.cnc                         # may be None

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def we_v(self):
        return self.__we_v


    @property
    def we_c(self):
        return self.__we_c


    @property
    def cnc(self):
        return self.__cnc


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "PIDDatum:{we_v:%0.6f, we_c:%s, cnc:%s}" % (self.we_v, self.we_c, self.cnc)



"""
Created on 19 Apr 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import requests

from abc import ABC
from http import HTTPStatus

from scs_core.aws.config.endpoint import APIEndpoint

from scs_core.client.http_exception import HTTPException

from scs_core.data.json import JSONify

from scs_core.sys.logging import Logging


# --------------------------------------------------------------------------------------------------------------------

class APIClient(ABC):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, reporter=None):
        """
        Constructor
        """
        self.__reporter = reporter
        self.__logger = Logging.getLogger()


    # ----------------------------------------------------------------------------------------------------------------

    def _auth_headers(self, auth=APIEndpoint.DEFAULT_AUTH):
        header = APIEndpoint.header(auth=auth)
        self.__logger.debug('header: %s' % header)

        return header


    def _token_headers(self, token):
        header = APIEndpoint.header(token=token)
        self.__logger.debug('header: %s' % header)

        return header


    def _check_response(self, response):
        status = HTTPStatus(response.status_code)
        self.__logger.debug('status: %s' % status)

        if status != HTTPStatus.OK:
            raise HTTPException.construct(status.value, response.reason, response.json())


    def _get_blocks(self, url, token, block_class, params=None, payload=None):
        self.__logger.debug("url: %s" % url)

        while True:
            response = requests.get(url, headers=self._token_headers(token), params=params, data=JSONify.dumps(payload))
            self._check_response(response)

            self.__logger.debug("response: %s" % response.json())

            # messages...
            block = block_class.construct_from_jdict(response.json())
            self.__logger.debug("block: %s" % block)

            for item in block.items:
                yield item

            # report...
            if self.__reporter:
                self.__reporter.print(len(block), block_start=block.start())

            # next request...
            if block.next_url is None:
                break

            params = block.next_params(params)


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def _reporter(self):
        return self.__reporter


    @property
    def _logger(self):
        return self.__logger


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return self.__class__.__name__ + ":{reporter:%s}" % self.__reporter



"""
Created on 9 Nov 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import ssl

import http.client

import urllib.parse

from socket import gaierror, timeout as timeout_error

from scs_core.client.http_exception import HTTPException
from scs_core.client.http_status import HTTPStatus
from scs_core.client.resource_unavailable_exception import ResourceUnavailableException


# --------------------------------------------------------------------------------------------------------------------

class HTTPClient(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        """
        Constructor
        """
        self.__conn = None
        self.__host = None


    # ----------------------------------------------------------------------------------------------------------------

    def connect(self, host, secure=True, verified=True, timeout=None):
        # print("connect: host: {}, timeout: {}".format(host, timeout), file=sys.stderr)

        self.__host = host

        if secure:
            # noinspection PyProtectedMember,PyUnresolvedReferences
            context = None if verified else ssl._create_unverified_context()

            if timeout is not None:
                self.__conn = http.client.HTTPSConnection(host, context=context, timeout=timeout)
            else:
                self.__conn = http.client.HTTPSConnection(host, context=context)

        else:
            if timeout is not None:
                self.__conn = http.client.HTTPConnection(host, timeout=timeout)
            else:
                self.__conn = http.client.HTTPConnection(host)


    def close(self):
        if self.__conn:
            self.__conn.close()


    # ----------------------------------------------------------------------------------------------------------------

    def get(self, path, payload, headers):
        # data...
        params = urllib.parse.urlencode(payload) if payload else None
        query = path + '?' + params if params else path

        # print("get: query: {}".format(query))
        # print("headers: %s " % headers)

        # request...
        response = self.__request("GET", query, None, headers)
        data = response.read()

        # error...
        if response.status != HTTPStatus.OK:
            raise HTTPException.construct_from_res(response, data)

        # print("response: %s " % data.decode())

        return data.decode()


    def post(self, path, payload, headers):
        # request...
        response = self.__request("POST", path, payload, headers)
        data = response.read()

        # error...
        if response.status != HTTPStatus.OK and response.status != HTTPStatus.CREATED:
            raise HTTPException.construct_from_res(response, data)

        return data.decode()


    def put(self, path, payload, headers):
        # request...
        response = self.__request("PUT", path, payload, headers)
        data = response.read()

        # error...
        if response.status != HTTPStatus.OK and response.status != HTTPStatus.NO_CONTENT:
            raise HTTPException.construct_from_res(response, data)

        return data.decode()


    def delete(self, path, headers):
        # request...
        response = self.__request("DELETE", path, "", headers)
        data = response.read()

        # error...
        if response.status != HTTPStatus.OK and response.status != HTTPStatus.NO_CONTENT:
            raise HTTPException.construct_from_res(response, data)

        return data.decode()


    # ----------------------------------------------------------------------------------------------------------------

    def __request(self, method, url, body, headers):
        try:
            self.__conn.request(method, url, body=body, headers=headers)
            return self.__conn.getresponse()

        except (gaierror, timeout_error, http.client.CannotSendRequest, OSError) as ex:
            self.__conn.close()

            raise ResourceUnavailableException(self.__host + url, ex)


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "HTTPClient:{conn:%s, host:%s}" % (self.__conn, self.__host)



"""
Created on 07 Apr 2021

@author: Jade Page (jade.page@southcoastscience.com)

https://stackoverflow.com/questions/36932/how-can-i-represent-an-enum-in-python
"""

import json

from collections import OrderedDict
from enum import Enum
from urllib.parse import parse_qs, urlparse

from scs_core.aws.client.api_intercourse import APIResponse

from scs_core.data.str import Str
from scs_core.data.datum import Datum

from scs_core.sample.configuration_sample import ConfigurationSample


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationRequest(object):
    """
    classdocs
    """

    class Mode(Enum):
        LATEST = 1
        HISTORY = 2
        DIFF = 3
        TAGS_ONLY = 4

    TAG_FILTER = 'tagFilter'
    EXACT_MATCH = 'exactMatch'
    RESPONSE_MODE = 'responseMode'
    EXCLUSIVE_START_KEY = 'exclusiveStartKey'

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        tag_filter = qsp.get(cls.TAG_FILTER)
        exact_match = Datum.is_true(qsp.get(cls.EXACT_MATCH))

        try:
            response_mode = cls.Mode[qsp.get(cls.RESPONSE_MODE)]
        except KeyError:
            response_mode = None

        esk_json = qsp.get(cls.EXCLUSIVE_START_KEY)
        exclusive_start_key = ExclusiveStartKey.construct_from_qsp(json.loads(esk_json)) if esk_json else None

        return cls(tag_filter, exact_match, response_mode, exclusive_start_key=exclusive_start_key)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag_filter, exact_match, response_mode, exclusive_start_key=None):
        """
        Constructor
        """
        self.__tag_filter = tag_filter                          # string
        self.__exact_match = bool(exact_match)                  # bool
        self.__response_mode = response_mode                    # Mode enum

        self.__exclusive_start_key = exclusive_start_key        # ExclusiveStartKey


    # ----------------------------------------------------------------------------------------------------------------

    def is_valid(self):
        if self.response_mode is None:
            return False

        return True


    def latest(self):
        return self.__response_mode == self.Mode.LATEST


    def history(self):
        return self.__response_mode == self.Mode.HISTORY


    def diff(self):
        return self.__response_mode == self.Mode.DIFF


    def tags_only(self):
        return self.__response_mode == self.Mode.TAGS_ONLY


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        params = {
            self.EXACT_MATCH: self.exact_match,
            self.RESPONSE_MODE: self.response_mode.name
        }

        if self.tag_filter is not None:
            params[self.TAG_FILTER] = self.tag_filter

        if self.exclusive_start_key is not None:
            params[self.EXCLUSIVE_START_KEY] = json.dumps(self.exclusive_start_key.params())

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def tag_filter(self):
        return self.__tag_filter


    @property
    def exact_match(self):
        return self.__exact_match


    @property
    def response_mode(self):
        return self.__response_mode


    @property
    def exclusive_start_key(self):
        return self.__exclusive_start_key


    @exclusive_start_key.setter
    def exclusive_start_key(self, exclusive_start_key):
        self.__exclusive_start_key = exclusive_start_key


    @property
    def exclusive_start_key_params(self):
        return self.__exclusive_start_key.params() if self.__exclusive_start_key else None


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationRequest:{tag_filter:%s, exact_match:%s, response_mode:%s, exclusive_start_key:%s}" % \
               (self.tag_filter, self.exact_match, self.response_mode, self.exclusive_start_key)


# --------------------------------------------------------------------------------------------------------------------

class ExclusiveStartKey(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_qsp(cls, qsp):
        if not qsp:
            return None

        rec = qsp.get('rec')
        tag = qsp.get('tag')

        return cls(rec, tag)


    @classmethod
    def construct_from_jdict(cls, jdict):
        return cls.construct_from_qsp(jdict)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, rec, tag):
        """
        Constructor
        """
        self.__rec = rec                    # string
        self.__tag = tag                    # string


    # ----------------------------------------------------------------------------------------------------------------

    def params(self):
        params = {'tag': self.tag}

        if self.rec is not None:
            params['rec'] = self.rec

        return params


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def rec(self):
        return self.__rec


    @property
    def tag(self):
        return self.__tag


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ExclusiveStartKey:{rec:%s, tag:%s}" %  (self.rec, self.tag)


# --------------------------------------------------------------------------------------------------------------------

class ConfigurationResponse(APIResponse):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict):
        if not jdict:
            return None

        mode = ConfigurationRequest.Mode[jdict.get('mode')]

        items = []
        if jdict.get('Items'):
            for item_jdict in jdict.get('Items'):
                item = item_jdict if mode == ConfigurationRequest.Mode.TAGS_ONLY else  \
                    ConfigurationSample.construct_from_jdict(item_jdict)
                items.append(item)

        next_url = jdict.get('next')

        return cls(mode, items, next_url=next_url)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, mode, items, next_url=None):
        """
        Constructor
        """
        self.__mode = mode                                  # ConfigurationRequest.Mode member
        self.__items = items                                # list of ConfigurationSample or string
        self.__next_url = next_url                          # URL string


    def __len__(self):
        return len(self.items)


    # ----------------------------------------------------------------------------------------------------------------

    def next_params(self, _):
        return parse_qs(urlparse(self.next_url).query)


    # ----------------------------------------------------------------------------------------------------------------

    def as_json(self):
        jdict = OrderedDict()

        if self.mode is not None:
            jdict['mode'] = self.mode.name

        if self.items is not None:
            jdict['Items'] = self.items
            jdict['itemCount'] = len(self.items)

        if self.next_url is not None:
            jdict['next'] = self.next_url

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def mode(self):
        return self.__mode


    @property
    def items(self):
        return self.__items


    @property
    def next_url(self):
        return self.__next_url


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "ConfigurationResponse:{mode:%s, items:%s, next_url:%s}" % \
               (self.mode, Str.collection(self.items), self.next_url)



"""
Created on 24 May 2023

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

document example:
{"key-id": "ABC", "secret-key": "123"}
"""

import json
import requests

from scs_core.aws.client.access_key import AccessKey
from scs_core.aws.client.api_client import APIClient
from scs_core.aws.config.endpoint import APIEndpoint


# --------------------------------------------------------------------------------------------------------------------

class Endpoint(APIEndpoint):
    @classmethod
    def configuration(cls):
        return cls('CogDevKeyAPI/CognitoDeviceKey',
                   'https://a0vjvahph1.execute-api.us-west-2.amazonaws.com/default/CognitoDeviceKey')


# --------------------------------------------------------------------------------------------------------------------

class AccessKeyManager(APIClient):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self):
        super().__init__()


    # ----------------------------------------------------------------------------------------------------------------

    def get(self, token):
        response = requests.get(Endpoint.url(), headers=self._token_headers(token))
        self._check_response(response)

        for id, secret in json.loads(response.json()).items():
            return AccessKey(id, secret)                                # only one key



"""
Created on 9 Jan 2024

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)
"""

import os
import pytz
import re

from datetime import datetime as dt

from scs_core.sys.filesystem import Filesystem, File


# --------------------------------------------------------------------------------------------------------------------

class CSVDataLog(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def directory_name(datetime):
        if datetime is None:
            raise ValueError("datetime may not be None")

        return "%04d-%02d" % (datetime.year, datetime.month)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, root_path, topic_subject, tag=None, timeline_start=None):
        """
        Constructor
        """
        self.__root_path = root_path                        # string
        self.__topic_subject = topic_subject                # string
        self.__tag = tag                                    # string

        self.__timeline_start = timeline_start              # datetime


    # ----------------------------------------------------------------------------------------------------------------

    def mkdir(self):
        Filesystem.mkdir(self.directory_path())


    def directory_path(self):
        return os.path.join(self.root_path, self.directory_name(self.timeline_start))


    def file_path(self):
        filename = CSVDataLogFile.name(self.timeline_start, self.topic_subject, self.tag)

        return os.path.join(self.directory_path(), filename)


    def in_timeline(self, localised_datetime):
        if self.timeline_start is None:
            return False

        utc_localised_datetime = localised_datetime.utc()

        return utc_localised_datetime.datetime.date() == self.__timeline_start.date()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def root_path(self):
        return self.__root_path


    @property
    def topic_subject(self):
        return self.__topic_subject


    @property
    def tag(self):
        return self.__tag


    @tag.setter
    def tag(self, tag):
        self.__tag = tag


    @property
    def timeline_start(self):
        return self.__timeline_start


    @timeline_start.setter
    def timeline_start(self, localised_timeline_start):
        self.__timeline_start = localised_timeline_start.utc_datetime


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVDataLog:{root_path:%s, tag:%s, topic_subject:%s, timeline_start:%s}" % \
               (self.root_path, self.tag, self.topic_subject, self.timeline_start)


# --------------------------------------------------------------------------------------------------------------------

class CSVDataLogFile(object):
    """
    classdocs
    """

    # ----------------------------------------------------------------------------------------------------------------

    @staticmethod
    def name(datetime, topic_subject, tag=None):
        if datetime is None:
            raise ValueError("datetime may not be None")

        if tag is None:
            return "%s-%4d-%02d-%02d-%02d-%02d-%02d.csv" % \
                   (topic_subject, datetime.year, datetime.month, datetime.day,
                    datetime.hour, datetime.minute, datetime.second)

        return "%s-%s-%4d-%02d-%02d-%02d-%02d-%02d.csv" % \
               (tag, topic_subject, datetime.year, datetime.month, datetime.day,
                datetime.hour, datetime.minute, datetime.second)


    @classmethod
    def construct(cls, file: File):
        match = re.match(r'^(.+-)?([^-]+)-(\d{4})-(\d{2})-(\d{2})-(\d{2})-(\d{2})-(\d{2})\.csv',
                         file.name)

        if not match:
            return None

        fields = match.groups()

        # fields...
        tag = None if fields[0] is None else fields[0][:-1]

        topic_subject = fields[1]

        year = int(fields[2])
        month = int(fields[3])
        day = int(fields[4])

        hour = int(fields[5])
        minute = int(fields[6])
        second = int(fields[7])

        created_datetime = dt(year, month=month, day=day, hour=hour, minute=minute, second=second,
                              tzinfo=pytz.UTC)

        return cls(created_datetime, topic_subject, tag, file)


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, created_datetime, topic_subject, tag, file: File):
        """
        Constructor
        """
        self.__created_datetime = created_datetime          # datetime (offset-aware to UTC)
        self.__topic_subject = topic_subject                # string
        self.__tag = tag                                    # string

        self.__file = file                                  # File


    # ----------------------------------------------------------------------------------------------------------------

    def path(self):
        return self.__file.path()


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def created_datetime(self):
        return self.__created_datetime


    @property
    def topic_subject(self):
        return self.__topic_subject


    @property
    def tag(self):
        return self.__tag


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        return "CSVDataLogFile:{created_datetime:%s, tag:%s, topic_subject:%s, file:%s}" % \
               (self.created_datetime, self.tag, self.topic_subject, self.__file)



"""
Created on 20 Oct 2016

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

example document:
{"rec": "2021-10-11T10:59:40Z", "tag": "scs-be2-3", "ver": 2.0, "src": "AFE",
"val": {"NO2": {"weV": 0.29057, "aeV": 0.29544, "weC": 0.00131, "cnc": 20.9, "vCal": 15.41},
"Ox": {"weV": 0.40101, "aeV": 0.39969, "weC": 0.00235, "cnc": 55.7, "vCal": 6.61, "xCal": -0.390731},
"CO": {"weV": 0.44069, "aeV": 0.30213, "weC": 0.16795, "cnc": 685.4, "vCal": 562.256},
"sht": {"hmd": 52.9, "tmp": 21.9}}, "exg": {"src": "vB20", "val": {"NO2": {"cnc": 19.3}}}}
"""

from collections import OrderedDict

from scs_core.climate.sht_datum import SHTDatum

from scs_core.data.datetime import LocalizedDatetime
from scs_core.data.str import Str

from scs_core.gas.a4.a4_calibrated_datum import A4CalibratedDatum
from scs_core.gas.afe.afe_datum import AFEDatum
from scs_core.gas.afe.pt1000_datum import Pt1000Datum
from scs_core.gas.pid.pid_calibrated_datum import PIDCalibratedDatum
from scs_core.gas.scd30.scd30_datum import SCD30Datum

from scs_core.sample.sample import Sample


# --------------------------------------------------------------------------------------------------------------------

class GasesSample(Sample):
    """
    classdocs
    """

    VERSION = 1.0

    __NON_ELECTROCHEM_FIELDS = ('pt1', 'sht', 'CO2', 'VOC', 'VOCe')
    __VOC_FIELDS = ('VOC', 'VOCe')

    # ----------------------------------------------------------------------------------------------------------------

    @classmethod
    def construct_from_jdict(cls, jdict, skeleton=False):
        if not jdict:
            return None

        # Sample...
        tag = jdict.get('tag')
        rec = LocalizedDatetime.construct_from_jdict(jdict.get('rec'))

        try:
            version = round(float(jdict.get('ver')), 1)
        except (TypeError, ValueError):
            version = cls.DEFAULT_VERSION

        src = jdict.get('src')
        val = jdict.get('val')
        exegeses = jdict.get('exg')

        # GasesSample...
        node = val.get('pt1')
        pt1000_datum = None if node is None else Pt1000Datum(node.get('v'), temp=node.get('tmp'))

        node = val.get('sht')
        sht_datum = SHTDatum(node.get('hmd'), node.get('tmp'))

        node = val.get('CO2')
        scd30_datum = None if node is None else SCD30Datum(node.get('cnc'), None, None)

        sns = OrderedDict()

        for field, node in val.items():
            if field not in cls.__NON_ELECTROCHEM_FIELDS:
                sns[field] = A4CalibratedDatum.construct_from_jdict(node)

            if field in cls.__VOC_FIELDS:
                sns[field] = PIDCalibratedDatum.construct_from_jdict(node)

        electrochem_datum = AFEDatum(pt1000_datum, *list(sns.items()))

        return cls(tag, rec, scd30_datum, electrochem_datum, sht_datum, version=version, src=src, exegeses=exegeses)


    @classmethod
    def has_invalid_value(cls):
        # TODO: implement has_invalid_value
        return False


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, tag, rec, scd30_datum, electrochem_datum, sht_datum, version=None, src=None, exegeses=None):
        """
        Constructor
        """
        if version is None:
            version = self.VERSION

        super().__init__(tag, rec, version, src=src, exegeses=exegeses)

        self.__scd30_datum = scd30_datum                            # SCD30Datum
        self.__electrochem_datum = electrochem_datum                # AFEDatum or ISIDatum
        self.__sht_datum = sht_datum                                # SHT31Datum


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def values(self):
        jdict = OrderedDict()

        if self.scd30_datum is not None:
            jdict['CO2'] = {'cnc': self.scd30_datum.co2}

        if self.electrochem_datum is not None:
            for key, value in self.electrochem_datum.sns.items():
                jdict[key] = value.as_json()

            try:
                if self.electrochem_datum.pt1000 is not None:
                    jdict['pt1'] = self.electrochem_datum.pt1000.as_json()
            except AttributeError:
                pass

        if self.sht_datum is not None:
            jdict['sht'] = self.sht_datum.as_json()

        return jdict


    # ----------------------------------------------------------------------------------------------------------------

    @property
    def scd30_datum(self):
        return self.__scd30_datum


    @property
    def electrochem_datum(self):
        return self.__electrochem_datum


    @property
    def sht_datum(self):
        return self.__sht_datum


    # ----------------------------------------------------------------------------------------------------------------

    def __str__(self, *args, **kwargs):
        exegeses = Str.collection(self.exegeses)

        return "GasesSample:{tag:%s, rec:%s, src:%s, exegeses:%s, " \
               "scd30_datum:%s, electrochem_datum:%s, sht_datum:%s}" % \
            (self.tag, self.rec, self.src, exegeses,
             self.scd30_datum, self.electrochem_datum, self.sht_datum)



"""
Created on 21 Mar 2019

@author: Bruno Beloff (bruno.beloff@southcoastscience.com)

Site information required for a device that may join the AirNow-I project

example JSON:
{"host": "sftp.airnowdata.org", "port": 22, "username": "UNEPdatauser", "password": "xxx",
"remote-path": "/UNEP/AQCSV"}
"""

from scs_core.client.sftp_client_conf import SFTPClientConf

from scs_core.data.json import PersistentJSONable


# --------------------------------------------------------------------------------------------------------------------

class AirNowUploaderConf(SFTPClientConf, PersistentJSONable):
    """
    classdocs
    """

    __FILENAME = "airnow_uploader_conf.json"

    @classmethod
    def persistence_location(cls):
        return cls.conf_dir(), cls.__FILENAME


    # ----------------------------------------------------------------------------------------------------------------

    def __init__(self, host, port, username, password, remote_path):
        super().__init__(host, port, username, password, remote_path)
